package F5::BdCfg::Bundle;

use strict;
use warnings;

use Getopt::Long;
use JSON;
use File::Basename;
use File::Find;
use File::Spec;
use File::Path qw(rmtree);
use Cwd qw(getcwd);
use YAML::XS;
use Digest::MD5;

use F5::ASMConfig::GuaranteedClient;
use F5::Deploy;
use F5::Utils::File;
use F5::PrepareFiles;
use F5::ConfigSet;
use F5::LockFactory;
use F5::SetActive;

my $BD_CONFIG_DIR          = F5::Deploy::get_path_by_type("bd_config");
my $BD_CONFIG_TABLES_DIR   = F5::Deploy::get_path_by_type("bd_config_tables");
my $DATABASE_BASE_DIR      = F5::Deploy::get_path_by_type("database");
my $POLICY_BUNDLE_BASE_DIR = F5::Deploy::get_path_by_type("var") . "/policy_bundles";
my $OPT                    = F5::Deploy::get_path_by_type("top_level");
my $GLOBAL_FILE            = "$POLICY_BUNDLE_BASE_DIR/global";
my $CATALOG_FILE           = "catalog.yaml";
my $VERSION                = File::Slurp::read_file("$OPT/VERSION.common", chomp => 1);
my $call_flags             = F5::ASMConfig::CallFlags->new( skip_set_active => 1, suppress_logging => 1, skip_set_sigs_active => 1 );

#----------------------------------------------------------------------

sub new {
    my ( $class, %params ) = @_;
    my $self = {};
    bless $self, $class;
    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my ( $self, %params ) = @_;

    $self->{workdir}    = getcwd;
    $self->{config_set} = F5::ConfigSet->new();
    $self->{dbh}        = $self->{config_set}->{dbh};
    $self->{logger}     = $params{logger} || F5::Logger->new('/ts/log/apcompile');
    $self->{global_state_init} = 0;

    my $global_cache_exists = $self->load_global_cache();
    # Backup initial databases
    my $overwrite_existing = $global_cache_exists ? 0 : 1;
    F5::DbUtils::copy_sqlite_databases( suffix => 'initial', overwrite_existing => $overwrite_existing );

    $self->{locker} = F5::LockFactory::get_locker( lock_filename => $F5::DbUtils::LOCK_FILE );
}

#----------------------------------------------------------------------

sub load_global_cache {
    my ($self, %params) = @_;
    my $global_file = $params{global_file} || $GLOBAL_FILE;
    if ( -e $global_file) {
        my $cache = YAML::XS::LoadFile($global_file);
        if (!$cache->{'global-state'}->{version} || $cache->{'global-state'}->{version} ne $VERSION) {
            $self->{logger}->info("Found stale cache: Version != $VERSION"); 
            return 0;
        }
        $self->{global_state} = $cache->{'global-state'};
        $self->{global_settings_md5} = $cache->{global_settings_md5};
    }
    return 1;
}

#----------------------------------------------------------------------

sub get_client {
    my ($self) = @_;
    if ( $self->{client} ) {
        return $self->{client};
    }
    require F5::ASMConfig::GuaranteedClient;
    $self->{client} = F5::ASMConfig::GuaranteedClient::get_client( username => 'import_export_policy', dbh => $self->{dbh} );
    return $self->{client};
}

#----------------------------------------------------------------------

sub load_global_state {
    my ( $self, %params ) = @_;
    my $input_filename = File::Spec->rel2abs( $params{filename} );

    die { "error_message" => "Missing global settings/state file - '$input_filename'" } unless ( -f $input_filename );

    my $is_json = F5::Utils::File::is_file_json( filename => $input_filename );
    my $is_tgz  = F5::Utils::File::check_filetype( filename => $input_filename, valid_types => ['gzip'] )->{completed_successfully};
    my $global_settings_md5 = F5::Utils::File::get_file_md5($input_filename);
    my $global_settings_file_changed =  (!$self->{global_settings_md5} || $self->{global_settings_md5} ne $global_settings_md5);
    $self->{logger}->info("Global settings file $input_filename has " . ($global_settings_file_changed ? '' : 'not ') . "changed");
    if ($is_json) {
        if ($global_settings_file_changed) {
            #only reset the DB if the file has changed. The rest of the work should still be done in case there are references (like user defined signatures) that have changed.
            $self->reset_global( skip_set_global_info => 1 );
        }
        require F5::GlobalSettings;
        eval {
            $self->{global_settings} = F5::GlobalSettings->new(
                                                               client   => $self->get_client(),
                                                               filename => $input_filename
                                                          );
        };
        if ($@) {
            my $err = $@;
            die { "error_message" => "Failed to load WAF settings file - $err" };
        }
        my $apply_rv = $self->{global_settings}->apply( call_flags => $call_flags );

        if ( defined $apply_rv->{completed_successfully} && !$apply_rv->{completed_successfully} ) {
            die { "error_message" => "Failed to apply WAF settings - " . $apply_rv->{failure_msg} };
        }
        $self->set_global_state_info(
            user_defined_signatures    => $apply_rv->{user_defined_signatures},
            cookie_seed                => $apply_rv->{cookie_protection},
            adaptive_violation_service => $apply_rv->{adaptive_violation_service},
            global_settings_md5        => $global_settings_md5,
        );
        return { completed_successfully => 1 };
    }
    elsif ($is_tgz && $global_settings_file_changed) {
        my $rv;
        my $unzip_dir = "$POLICY_BUNDLE_BASE_DIR/load_global.$$";
        F5::Utils::File::make_path($unzip_dir);
        F5::SystemUtils::run_system_tool( command_line_params => [ 'tar', '-zpx', '-f', $input_filename, '-C', $unzip_dir, ] );
        my $global_file = "$unzip_dir/global";
        if (! -e $global_file) {
            die { "error_message" => "Invalid global file format" };
        }
        if ($self->load_global_cache(global_file => $global_file)) {
            find (
                {
                 wanted => sub {
                        if ( -f $File::Find::name ) {
                            my $actual_file_location = $File::Find::name;
                            if ( $_ eq "global" ) {
                                $actual_file_location = $GLOBAL_FILE;
                            }
                            else {
                                $actual_file_location =~ s/$unzip_dir/$OPT/;
                            }
                            F5::Utils::File::copy_files_recursively( from_dir => $File::Find::name, to_dir => $actual_file_location );
                        }
                    },
                },
                $unzip_dir
            );
            $self->set_global_state_info(global_settings_md5 => $global_settings_md5);
            $rv = { completed_successfully => 1 };
        }
        else {
            die { "error_message" => "Incorrect version for global file: Version != $VERSION" };
        }
        rmtree($unzip_dir);
        return $rv;
    }
    else {
        die { "error_message" => "Invalid global file format" };
    }
}

#----------------------------------------------------------------------

my $POLICY_SOURCE_FILENAME = "policy.json";

sub create_policy_bundle {
    my ( $self, %params ) = @_;
    my $bundle_policy;
    my $bundle_policy_name;
    my $policy_name;
    my $policy_filename    = File::Spec->rel2abs( $params{policy_filename} );
    my $output_filename    = File::Spec->rel2abs( $params{output_file} );
    my $include_defaults = defined $params{include_defaults} ? $params{include_defaults} : 1;
    my $include_source   = $params{include_source};

    my $policies_to_bundle;
    if ($include_defaults) {
        $policies_to_bundle  = $self->{config_set}->get_default_policy_files();
    }

    die { "error_message" => "Missing policy file - '$policy_filename'" } unless ( -f $policy_filename );

    $self->set_global_state_info();

    if ( exists $policies_to_bundle->{$policy_filename} ) {
        $bundle_policy_name = $policies_to_bundle->{$policy_filename};
    }
    else {
        $bundle_policy_name = $policies_to_bundle->{$policy_filename} = $self->extract_policy_name( filename => $policy_filename );
    }

    my %catalog_policies = (
        bundlePolicyName => $bundle_policy_name,
        policies         => {},
    );

    my $cached_config;
    my $save_cache = 0;
    my $bundle_policy_has_references = 0;
    my $cached_config_set = F5::DbUtils::select_singleton(dbh => $self->{dbh}, table_name => 'PLC.CONFIG_SET');
    if ($cached_config_set->{config_set}) {
        $cached_config = YAML::XS::Load($cached_config_set->{config_set})
    }

    for my $policy_filename ( keys %$policies_to_bundle ) {
        my $policy;
        my $policy_name = $policies_to_bundle->{$policy_filename};
        
        if ($include_defaults && $policies_to_bundle->{$policy_filename} eq $bundle_policy_name && ( $policy_name eq "app_protect_default_policy" || $policy_name eq "app_protect_strict_policy" ) &&
           (!defined $ENV{APP_PROTECT_GENERATE_SECURITY_DEFAULTS})) {
                die { "error_message" => "Policy '$policy_name' uses a reserved name and may not be compiled to a custom policy bundle" }
        }
        my $policy_file_md5 = F5::Utils::File::get_file_md5($policy_filename);
        my $do_import = 1;
        my $do_apply = 1;
        if ($cached_config && $cached_config->{policies}->{$policy_name}) {
            $policy = F5::ASMConfig::Entity::Policy->new(dbh => $self->{dbh}, get_criteria => { policy_name => $policy_name });
            $do_import = !$policy || !$self->{config_set}->is_policy_unmodified(previous_policy => $cached_config->{policies}->{$policy_name},
                                                                                policy => { import_filename_md5 => $policy_file_md5 });
            $do_apply = $do_import || !$policy->{account_id};
            $save_cache ||= $do_import;
            if ( $policies_to_bundle->{$policy_filename} eq $bundle_policy_name ) {
                 $bundle_policy_has_references = $cached_config->{policies}->{$policy_name}->{external_reference_count};
            }
        }

        my @policy_warnings;
        if ($do_import) {
            my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(
                filename              => $policy_filename,
                force_policy_name_flg => 1,
                policy_name           => $policy_name,
            );
            my $import_rv;

            eval {
                $import_rv       = $self->get_client->import_policy( $self->get_client->{session}, $call_flags, $import_export_properties );
                $policy_name     = $import_rv->{policy_name};
                if ( $import_rv && $import_rv->{msgs} ) {
                    @policy_warnings = map { $_->{msg} } grep { $_->{level} eq 'warning' } @{ $import_rv->{msgs} };
                    if (@policy_warnings) {
                        $self->{warnings} = \@policy_warnings;    # TODO: merge warnings from different policies
                    }
                }
            };

            if ( !$import_rv || !$import_rv->{completed_successfully} ) {
                my $err = $@;
                if ( $import_rv && $import_rv->{msgs} ) {
                    $err = join( "\n", map { $_->{msg} } @{ $import_rv->{msgs} } );
                }
                elsif ( ref $err ) {
                    $err = $err->{display_message} || $err->{error_message};
                }
                die { "error_message" => "Failed to import policy from '$policy_filename - $err'" };
            }
            $policy = $import_rv->{policy};
            if ($cached_config && $cached_config->{policies}->{$policy_name}) {
                $cached_config->{policies}->{$policy_name}->{external_reference_count} = $import_rv->{external_reference_count};
                $cached_config->{policies}->{$policy_name}->{external_references} = $import_rv->{external_references};
                $cached_config->{policies}->{$policy_name}->{import_filename_md5} = $policy_file_md5;
            }
            if ( $policies_to_bundle->{$policy_filename} eq $bundle_policy_name ) {
                 $bundle_policy_has_references = $import_rv->{external_reference_count};
            }
        }

        if ($do_apply) {
            my $set_active_result = F5::SetActive::set_active(
                policy_id              => $policy->{policy_id},
                expanded_return_value  => 1,
                do_save_policy_history => 0,
                do_update_core         => 0,
                dbh                    => $self->{dbh},
                wait_for_enforcer      => 0,
                use_daemon             => 0,
                update_ramcache        => 0,
                asm_config_handler     => $self->get_client(),
                send_sync              => 0,
            );
            die { "error_message" => "Failed to compile policy - '$policy_name'" } unless $set_active_result->{completed_successfully};
        }

        my $rv = $self->{client}->get_policy_adaptive_violation_rating($self->get_client->{session}, $call_flags, $policy);
        my $adaptiveViolationRating = {enabled => $rv->{flg_enabled} ? JSON::true : JSON::false};

        $catalog_policies{"policies"}{$policy_name} = {
            originalFilename => $policy_filename,
            policyNameCrc    => $policy->{policy_name_crc},
            fileMD5          => $policy_file_md5,
            'adaptive-violation-rating' => $adaptiveViolationRating,
        };
        $catalog_policies{"policies"}{$policy_name}{"warnings"} = \@policy_warnings if @policy_warnings;

        if ( $policies_to_bundle->{$policy_filename} eq $bundle_policy_name ) {
            $bundle_policy = $policy;
            if ($include_source) {
                $catalog_policies{"policies"}{$policy_name}{"sourceFilenameInBundle"} = $POLICY_SOURCE_FILENAME;
            }
        }
    }

    my $set_active_result = F5::SetActive::set_active(
        asm_config_handler    => $self->get_client(),
        dbh                   => $self->{dbh},
        do_update_core        => 1,
        expanded_return_value => 1,
        send_sync             => 0,
        wait_for_enforcer     => 0,
        publish_manifest      => 0,
    );

    die { "error_message" => "Failed to prepare bundle files" } unless $set_active_result->{completed_successfully};

    my $latest_full_dir = F5::PrepareFiles::find_latest_full_dir();
    my $bundle_create_dir = dirname($latest_full_dir);
    my $global_hash     = $self->{global_state}->{hash};

    my $subbundle_root_tmp        = "$POLICY_BUNDLE_BASE_DIR/save_policy__$bundle_policy_name.$$";
    my $subbundle_global          = "$subbundle_root_tmp/subbundle/global/$global_hash";
    my $subbundle_data_protection = "$subbundle_global/files/data_protection/";
    my $gen_path                  = $latest_full_dir =~ s/$BD_CONFIG_DIR//r;
    F5::Utils::File::make_path( $subbundle_global . $gen_path, $subbundle_data_protection );

    my $bd_cfg_manifest   = YAML::XS::LoadFile("$bundle_create_dir/bd_cfg_manifest.yaml");
    my $latest_full_files = $bd_cfg_manifest->{generations}->[0]->{fullFiles};

    my $crc_map;
    for my $policy ( keys %{ $catalog_policies{"policies"} } ) {
        my $crc = $catalog_policies{"policies"}{$policy}{"policyNameCrc"};
        $crc_map->{$crc}->{subbundle_policy} = "$subbundle_root_tmp/subbundle/policies/$policy";
        F5::Utils::File::make_path( $crc_map->{$crc}->{subbundle_policy} . $gen_path );
    }

    for my $file ( @{$latest_full_files} ) {
        my $file_policy_name_crc;
        if ( $file->{filename} =~ /__logprofile_\d+|vs_log_profile/ ) {
            next;
        }
        if ( $file->{filename} =~ /__account_(\d+)/ ) {
            $file_policy_name_crc = $1;
            if ( exists $crc_map->{$file_policy_name_crc} ) {
                File::Copy::cp( $file->{filename}, $crc_map->{$file_policy_name_crc}->{subbundle_policy} . $gen_path );
            }
            else {
                next;
            }
        }
        else {
            File::Copy::cp( $file->{filename}, $subbundle_global . $gen_path );
            undef $file_policy_name_crc;
        }
        if ( defined $file->{dependentFiles} ) {
            for my $dependency ( @{ $file->{dependentFiles} } ) {
                my $target_dir;
                if ( $dependency->{filename} =~ /data_protection/ ) {
                    $target_dir = $subbundle_data_protection;
                }
                elsif ( $dependency->{filename} =~ /bloom/ ) {
                    next;
                }
                elsif ( defined $file_policy_name_crc ) {
                    my $basedir = $dependency->{filename};
                    $basedir =~ s/$BD_CONFIG_DIR//;
                    $target_dir = $crc_map->{$file_policy_name_crc}->{subbundle_policy} . $basedir;
                }
                else {
                    # in case of new non-crc dependent files
                    $target_dir = "$subbundle_global/files/";
                }
                F5::Utils::File::copy_files_recursively( from_dir => $dependency->{filename}, to_dir => $target_dir );
            }
        }
    }

    if ($include_source) {
        my $policy_source_location = "$subbundle_root_tmp/$POLICY_SOURCE_FILENAME";
        if ($bundle_policy_has_references) {
            # Policy must be exported to be complete without external references
           $self->get_client->export_policy( $self->get_client->{session}, $call_flags, $bundle_policy, F5::ASMConfig::ImportExportProperties->new(
               filename => $policy_source_location,
               file_format => "json",
           ));
        }
        else {
            File::Copy::cp($policy_filename, $policy_source_location);
        }
    }

    my %catalog = ( %catalog_policies, 'global-state' => $self->{global_state} );
    $catalog{"timestamp"}       = F5::DateTimeUtils::convert_epoch_to_rfc3339( time() );
    $catalog{"softwareVersion"} = delete $catalog{'global-state'}->{version} || $VERSION;
    $YAML::XS::QuoteNumericStrings = 0;
    local $YAML::XS::Boolean = "JSON::PP";
    YAML::XS::DumpFile( "$subbundle_root_tmp/$CATALOG_FILE", \%catalog );

    my $rv = $self->archive_bundle(
        source_dir      => $subbundle_root_tmp,
        output_filename => $output_filename,
        source_file     => $include_source ? $POLICY_SOURCE_FILENAME : undef,
    );
    rmtree($subbundle_root_tmp);
    rmtree($bundle_create_dir);
    if (!$cached_config || !$cached_config->{policies}->{$bundle_policy_name}) {
        $self->delete_policy(policy => $bundle_policy);
    }
    if ($save_cache) {
        F5::DbUtils::upsert_singleton(dbh => $self->{dbh}, table_name => "PLC.CONFIG_SET",
           rec => {
             config_set => YAML::XS::Dump($cached_config),
             last_updated => time(),
           });
    }
    return $rv;
}

#----------------------------------------------------------------------

my $LOGGING_PROFILE_SOURCE_FILENAME = "logging_profile.json";

sub create_logging_profile_bundle {
    my ( $self, %params ) = @_;
    my $logging_profile_filename = File::Spec->rel2abs( $params{logging_profile_filename} );
    my $output_filename          = File::Spec->rel2abs( $params{output_file} );
    my $include_source           = $params{include_source};

    die { "error_message" => "Missing logging profile file - '$logging_profile_filename'" } unless ( -f $logging_profile_filename );
    my $logging_profile;
    $logging_profile = F5::Utils::File::load_file(
        filename  => $logging_profile_filename,
        json_only => 1
    );
    if (!$logging_profile->{content} && !$logging_profile->{filter}) {
         die { "error_message" => "No logging profile configuration found in '$logging_profile_filename'" };
    }
    my $logging_profile_name;
    if ( $logging_profile->{name} ) {
        $logging_profile_name = $logging_profile->{name};
    }
    else {
        $logging_profile_name = fileparse( $logging_profile_filename, '\..*' );
    }
    my $validate_chars = "-a-zA-Z0-9_";
    if ( $logging_profile_name =~ /([^$validate_chars])/ ) {
        die { "error_message" => "The value for logging profile name contains an illegal character (\"$1\")" };
    }
    my $logging_config = {
        logging => {
            filter  => $logging_profile->{filter}->{request_type},
            content => $logging_profile->{content},
        },
    };

    my %logging_profiles;
    $self->{config_set}->build_logging_config(
        policy           => { vs_info => { $logging_profile_name => $logging_config } },
        logging_profiles => \%logging_profiles
    );

    # take the first element, which is our log_profile config md5_hex
    my ($logging_profile_config_md5_hex) = ( keys %logging_profiles );

    require F5::CRC;
    my $logging_profile_name_crc = F5::CRC::get_crc32($logging_profile_config_md5_hex);
    my @logging_profiles_list    = values %logging_profiles;
    my $prepare_result           = $self->{config_set}->prepare_all_logging_profiles( logging_profiles => \@logging_profiles_list );
    my $set_active_result        = F5::SetActive::set_active(
        dbh                   => $self->{dbh},
        do_update_core        => 1,
        asm_config_handler    => $self->get_client(),
        expanded_return_value => 1,
        send_sync             => 0,
        publish_manifest      => 0,
    );

    my $latest_full_dir = F5::PrepareFiles::find_latest_full_dir();
    my @full_files      = F5::PrepareFiles::get_full_files_by_dir(
        base_dir             => $latest_full_dir,
        log_profile_name_crc => $logging_profile_name_crc,
    );

    die { "error_message" => "Failed to compile logging profile - '$logging_profile_name'" }
      unless $set_active_result->{completed_successfully};

    my $logging_profile_label     = "$logging_profile_name" . "_$logging_profile_config_md5_hex";
    my $subbundle_root_tmp        = "$POLICY_BUNDLE_BASE_DIR/save_logging_profile__$logging_profile_label.$$";
    my $gen_path                  = $latest_full_dir =~ s/$BD_CONFIG_DIR//r;
    my $subbundle_logging_profile = "$subbundle_root_tmp/subbundle/logging_profiles/$logging_profile_label/$gen_path";
    F5::Utils::File::make_path($subbundle_logging_profile);
    my $logging_profile_file_md5 = F5::Utils::File::get_file_md5($logging_profile_filename);

    my %catalog = (
        logging_profiles => {
            $logging_profile_label => {
                originalFilename => $logging_profile_filename,
                fileMD5          => $logging_profile_file_md5,
                logProfileCrc    => $logging_profile_name_crc,
            },
        },
    );

    if ($include_source) {
        File::Copy::cp($logging_profile_filename, "$subbundle_root_tmp/$LOGGING_PROFILE_SOURCE_FILENAME");
        $catalog{logging_profiles}{$logging_profile_label}{sourceFilenameInBundle} = $LOGGING_PROFILE_SOURCE_FILENAME;
    }

    $catalog{"softwareVersion"} = $VERSION;
    $catalog{"timestamp"}       = F5::DateTimeUtils::convert_epoch_to_rfc3339( time() );

    YAML::XS::DumpFile( "$subbundle_root_tmp/$CATALOG_FILE", \%catalog );

    for my $filename (@full_files) {
        File::Copy::cp( $filename, $subbundle_logging_profile ) if $filename =~ /logprofile_$logging_profile_name_crc/;
    }

    my $rv = $self->archive_bundle(
        source_dir      => $subbundle_root_tmp,
        output_filename => $output_filename,
        source_file     => $include_source ? $LOGGING_PROFILE_SOURCE_FILENAME : undef,
    );

    my $bundle_create_dir = dirname($latest_full_dir);
    rmtree($subbundle_root_tmp);
    rmtree($bundle_create_dir);
    return $rv;
}

#----------------------------------------------------------------------

sub update_global_settings {
    my ( $self, %params ) = @_;
    require F5::GlobalSettings;
    my $seed            = $params{cookie_seed} || F5::GlobalSettings::generate_cookie_seed_string();
    my $global_settings = F5::GlobalSettings->new(
        %params,
        cookieProtection => { seed => $seed },
        client                  => $self->get_client()
    );
    $global_settings->apply( call_flags => $call_flags );
    return $seed;
}

#----------------------------------------------------------------------

sub set_global_state_info {
    my ( $self, %params ) = @_;

    return if $self->{global_state_init};
    $self->{global_state_init} = 1;
    my $global_settings_md5 = $params{global_settings_md5};

    my $user_defined_signatures   = $params{user_defined_signatures} || ();
    my $adaptive_violation_service = $params{adaptive_violation_service};
    my $cookie_seed = $params{cookie_seed};

    my $md5_calc = Digest::MD5->new;

    if ( !$self->{global_state} || !%{$self->{global_state}} || $global_settings_md5 ) {
        if (!$global_settings_md5) {
            $self->{config_set}->reinstall_all_updates( call_flags => $call_flags );
        }
        F5::Utils::File::make_path($POLICY_BUNDLE_BASE_DIR);
        if ($cookie_seed || !defined($self->{global_state}->{"cookie-protection"})) {
            $cookie_seed ||= $self->update_global_settings(%params);
            $self->{global_state}->{"cookie-protection"} = { "seed" => $cookie_seed };
        }
        $self->{global_state}->{"user-defined-signatures"} = $user_defined_signatures if $user_defined_signatures;
        if ($adaptive_violation_service) {
            $self->{global_state}->{"adaptive-violation-rating"}->{"training-service"} = $adaptive_violation_service;
        }
    }

    if ($self->{global_state}->{'adaptive-violation-rating'} && $self->{global_state}->{'adaptive-violation-rating'}->{'training-service'}) {
        $md5_calc->add($self->{global_state}->{'adaptive-violation-rating'}->{'training-service'}->{host} || '');
        $md5_calc->add($self->{global_state}->{'adaptive-violation-rating'}->{'training-service'}->{secret} || '');
    }

    my $attack_signatures_version = $self->{config_set}->get_update_version( update_type => "signatures", generate_version => 1 );
    if ( %{$attack_signatures_version} ) {
        $attack_signatures_version->{revisionDatetime} = delete $attack_signatures_version->{revision_datetime};
        $self->{global_state}->{"attack-signatures"} = $attack_signatures_version;
        $md5_calc->add( $attack_signatures_version->{revisionDatetime} );
    }
    else {
        delete $self->{global_state}->{"attack-signatures"};
    }

    my $threat_campaigns_version = $self->{config_set}->get_update_version( update_type => "threat_campaigns", generate_version => 1 );
    if ( %{$threat_campaigns_version} ) {
        $threat_campaigns_version->{revisionDatetime} = delete $threat_campaigns_version->{revision_datetime};
        $self->{global_state}->{"threat-campaigns"} = $threat_campaigns_version;
        $md5_calc->add( $threat_campaigns_version->{revisionDatetime} );
    }
    else {
        delete $self->{global_state}->{"threat-campaigns"};
    }
 
    if ( $self->{global_state}->{"user-defined-signatures"} ) {
        for my $uds_file ( sort { $a->{fileMD5} cmp $b->{fileMD5} } @{ $self->{global_state}->{"user-defined-signatures"} } ) {
            $md5_calc->add( $uds_file->{fileMD5} );
        }
    }

    $self->{global_state}->{version} = $VERSION;
    $md5_calc->add( $self->{global_state}->{"cookie-protection"}->{"seed"} );
    $self->{global_state}->{hash} = $md5_calc->hexdigest();
    YAML::XS::DumpFile( $GLOBAL_FILE, { 'global-state' => $self->{global_state},
                                        'global_settings_md5' => $global_settings_md5 || $self->{global_settings_md5},
                                      } );
}

#----------------------------------------------------------------------

sub get_threat_campaigns_version {
    my ($self, %params) = @_;
    return $self->{global_state}->{'threat-campaigns'};
}

#----------------------------------------------------------------------

sub get_attack_signatures_version {
    my ($self, %params) = @_;
    return $self->{global_state}->{'attack-signatures'};
}

#----------------------------------------------------------------------


sub export_global_state {
    my ( $self, %params ) = @_;
    my $filesize;
    my $output_filename = File::Spec->rel2abs( $params{output_file} );

    $self->set_global_state_info();

    my $global_root_tmp = "$POLICY_BUNDLE_BASE_DIR/save_global.$$";
    my $global_db       = "$global_root_tmp/db";
    F5::Utils::File::make_path($global_root_tmp);
    F5::Utils::File::make_path($global_db);
    File::Copy::cp( "$DATABASE_BASE_DIR/PLC", $global_db );
    File::Copy::cp( "$DATABASE_BASE_DIR/DCC", $global_db );
    File::Copy::cp( $GLOBAL_FILE,             $global_root_tmp );

    my $archive_result =
      F5::SystemUtils::run_system_tool(
        command_line_params => [ 'tar', '-zpc', '-f', $output_filename, '-C', $global_root_tmp, 'db', 'global' ] );
    if ( $archive_result->{completed_successfully} ) {
        $filesize = -s $output_filename;
    }
    else {
        die { "error_message" => "Failed to save '$output_filename' - " . $archive_result->{standard_error} };
    }

    rmtree($global_root_tmp);

    return { completed_successfully => 1, filename => $output_filename, file_size => $filesize };
}

#----------------------------------------------------------------------

sub dump_bundle {
    my ( $self, %params ) = @_;
    my $input_filename = File::Spec->rel2abs( $params{filename}, $self->{workdir} );

    die { "error_message" => "Missing bundle file - '$input_filename'" } unless -f $input_filename;

    my $dump_catalog = F5::SystemUtils::run_system_tool( command_line_params => [ 'tar', '-O', '-xf', $input_filename, $CATALOG_FILE, ] );
    my $dump_global  = F5::SystemUtils::run_system_tool( command_line_params => [ 'tar', '-O', '-xf', $input_filename, 'global', ] );
    unless ( $dump_catalog->{completed_successfully} || $dump_global->{completed_successfully} ) {
        die { "error_message" => "Failed to open '$input_filename' - " . $dump_catalog->{standard_error} };
    }

    my $res = $dump_catalog->{standard_output} || $dump_global->{standard_output};

    return $res;

}

#----------------------------------------------------------------------

sub reset_global {
    my ( $self, %params ) = @_;

    F5::DbUtils::restore_sqlite_databases( suffix => 'initial' );

    #re-init db handles
    $self->init(logger => $self->{logger});

    # reinstall signatures updates
    my $install_rv = $self->{config_set}->reinstall_all_updates( call_flags => $call_flags );
    if ($install_rv->{prepare_sigs}) {
        require F5::NegativeSignatures::SigFile::Update;
        my $negsig_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $self->{dbh}, logger => $self->{logger});
        $negsig_update->set_sigs_active(asm_config_handler => $self->get_client,
                                        skip_set_active => 1,);
        F5::DbUtils::copy_sqlite_databases( suffix => 'initial', overwrite_existing => 1 );
    }

    # clean existing global_state_info
    unlink glob "$POLICY_BUNDLE_BASE_DIR/*";
    delete $self->{global_state};

    $self->set_global_state_info() unless $params{skip_set_global_info};

    # TODO: clean old bd_config files
}

#----------------------------------------------------------------------

sub extract_policy_name {
    my ( $self, %params ) = @_;
    my $filename = $params{filename};
    my $policy;
    require F5::YAML::Schema::Utils;
    if ( F5::Utils::File::is_file_json( filename => $filename ) ) {
        my $rv = F5::YAML::Schema::Utils::parse_json(filename => $params{filename});
        if ($rv->{completed_successfully}) {
            $policy = $rv->{data};
        }
        else {
            die $rv;
        }
    }
    if ( !ref $policy ) {
        die { "error_message" => "Bad file format - expected JSON" };
    }
    if ( !defined( $policy->{policy}->{name} ) ) {
        die { "error_message" => "Invalid file - expected declarative policy" };
    }
    return $policy->{policy}->{name};
}

#----------------------------------------------------------------------

sub archive_bundle {
    my ( $self, %params ) = @_;
    my $filesize;
    my $source_dir      = $params{source_dir};
    my $output_filename = $params{output_filename};
    my $source_file     = $params{source_file};

    my @command_line_params = ( 'tar', '-zpc', '-f', $output_filename, '-C', $source_dir, $CATALOG_FILE, 'subbundle' );
    if ($source_file) {
        push(@command_line_params, $source_file);
    }

    my $archive_result = F5::SystemUtils::run_system_tool( command_line_params => \@command_line_params );
    if ( $archive_result->{completed_successfully} ) {
        $filesize = -s $output_filename;
    }
    else {
        die { "error_message" => "Unable to write to '$output_filename'"};
    }

    return { completed_successfully => 1, filename => $output_filename, file_size => $filesize };
}

#----------------------------------------------------------------------

sub delete_policy {
    my ( $self, %params ) = @_;
    my $policy = $params{policy};
    my $client = $self->get_client;

    F5::SetActive::set_active(
        dbh                    => $self->{dbh},
        wait_for_enforcer      => 0,
        use_daemon             => 0,
        policy_id              => $policy->{policy_id},
        update_ramcache        => 0,
        do_update_core         => 0,
        do_save_policy_history => 0,
        deactivate_policy      => 1,
        publish_manifest       => 0,
        send_sync              => 0,
        asm_config_handler     => $client,
    );

    $client->delete_policy( $client->{session}, $call_flags, $policy );
    F5::DbUtils::vacuum_databases( $client->{dbh} );
}

1;
