package F5::BotSignatureFile::Update;

use strict;
use warnings;

use base qw(F5::Sigfile::Update);

use File::Path;

# Used to call gpg externally
use IPC::Open3;
use Symbol;
use File::Spec;

use F5::BotSignatures;
use F5::BigipUtils;
use F5::Deploy;
use F5::Provisioning;
use F5::Syslog;
use JSON;

my $TEMP_BASE_DIR = F5::Deploy::map_path("/ts/var/tmp");
our $SIGFILE_TYPE = "Bot Signature";
my $GPG_CMD = '/usr/bin/gpg';
my $IM_FILE_OUTPUT = 'ISO 9660 CD-ROM';
our $SAVE_DIR = F5::Deploy::map_path("/ts/var/bot_signatures");

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;
    $self->SUPER::init(%params);
    $self->{temp_base_dir} = $TEMP_BASE_DIR;
    $self->{sigfile_type} = $SIGFILE_TYPE;
    $self->{save_dir} = $SAVE_DIR;
    $self->{gpg_sign} = (exists $params{gpg_sign} ? $params{gpg_sign} : 1);
    $self->{default_rv} = { bot_signatures_count => 0,
                            bot_signatures_updated_count => 0,
                          };
}

#----------------------------------------------------------------------

sub apply_changes {
    my($self, %params) = @_;

    my $asm_config_handler = $params{asm_config_handler};
    my $username = $params{username};
    my $check_interval = $params{check_interval};
    my $update_mode = $params{update_mode};
    my $sigfile_status_update_mode = $params{sigfile_status_update_mode};
    my $flg_is_genesis_file = $params{flg_is_genesis_file};
    my $bot_signature_file_format = $params{bot_signature_file_format};
    my $skip_set_active = $params{skip_set_active};
    my $skip_set_sigs_active = $params{skip_set_sigs_active};
    my $gpg_sign = $params{gpg_sign};

    my $dbh = $self->{dbh};
    my $logger = $self->{logger};

    $self->{filename} = $params{filename};
    if ($self->{filename}) {
    if (! -e $self->{filename}) {
      F5::Syslog::ts_syslog( 'warning',
                             "Bot Signature file not found (%s)",
                             $self->{filename}
                           );

      $self->{filename} = '';
    }
    else {
        $self->unzip;
    }
  }

    my $bot_signature_file;
    my $bot_signature_file_update_status;
    eval {
        my $dir = "$self->{base_dir}/botsigs";
        $bot_signature_file_update_status = F5::BotSignatures::load_from_xml(
                                                                             dir => $dir,
                                                                             dbh => $dbh,
                                                                             logger => $logger,
                                                                            );

        $logger->debug("BotSignatures loaded to database") if $logger;

        rmtree($self->{base_dir});
    };

    if ($@) {
        my $error = $@;

        if (ref($error)) {
            die $error;
        }
        else {
            $error =~ s|\s+ at \s+ /\S+ \s+ line \s+ \d .*||sx;
            $self->store_error_and_die($error, \%params);
        }
    }

    if(!$params{dry_run} && !$skip_set_sigs_active) {
	      $logger->debug("Applying active policy for all accounts") if $logger;
        # Since bot signatures are included in negsig ACY and prepare_negsig_collections,
        # they are applied via signatures
        my $negsig_update = F5::NegativeSignatures::SigFile::Update->new(dbh => $dbh);
        $negsig_update->set_sigs_active(asm_config_handler => $asm_config_handler, skip_set_active => $skip_set_active);
    }

    my $success_msg = "$SIGFILE_TYPE updated successfully - $bot_signature_file_update_status->{bot_signatures_count} bot_signatures loaded, ($bot_signature_file_update_status->{bot_signatures_added_count} added, $bot_signature_file_update_status->{bot_signatures_updated_count} updated)";

    $bot_signature_file_update_status->{msg} = $success_msg;
    $bot_signature_file_update_status->{error_msg} = '';
    return ($bot_signature_file_update_status);
}

#----------------------------------------------------------------------

sub unzip {
  my($self) = @_;

  my $decrypted_filename;
  my $decrypted_is_temporary = 0;

  if ($self->{gpg_sign}) {
    $decrypted_is_temporary = 1;
    my $keyring_path = F5::Deploy::map_path("/ts/share/negsig/gpg_asm_sigfile_installer");
    $decrypted_filename = "$self->{temp_base_dir}/sigfile.tmp.$$.decrypted_sigfile.tgz";

    open(my $null1, ">", File::Spec->devnull);
    open(my $null2, File::Spec->devnull);
    my $gpg_err_fh = gensym;
    my $gpg_pid = open3($null1, $null2, $gpg_err_fh, 
                        $GPG_CMD, qw(--always-trust --yes --homedir), $keyring_path,
                        '--output', $decrypted_filename,
                        '--decrypt', $self->{filename});
    my $gpg_err = join('', <$gpg_err_fh>);
    waitpid($gpg_pid, 0);
    if ($?) {
      die "Bot Signature file is not valid\n";
    }
  }
  else {
    $decrypted_filename = $self->{filename};
  }

  $self->{base_dir} = "$self->{temp_base_dir}/sigfile.tmp.$$";
  $self->{base_dir_is_temporary} = 1;
  mkdir $self->{base_dir};
  my $sys_rv = F5::SystemUtils::run_system_tool(command_line_params => ['tar','-zmxf', $decrypted_filename, '-C', $self->{base_dir}],
                                                find_command_path => 1);
  unlink $decrypted_filename if $decrypted_is_temporary;

  if (!$sys_rv->{completed_successfully}) {
       die "Failed to unpack $self->{filename}: '$sys_rv->{standard_error}'\n";
  }
}

#----------------------------------------------------------------------

sub check_service_agreement {
    my($self, %params) = @_;

    if (! F5::Provisioning::is_asm_on()) {
        $self->store_error_and_die("ASM is not enabled on the device");
    }
    if (! F5::BigipUtils::is_feature_enabled('bot_defense')) {
        $self->store_error_and_die("Bot defense feature is not enabled");
    }
        
    return $self->SUPER::check_service_agreement(%params);
}


#----------------------------------------------------------------------

1;
