package F5::MBIP::Sidecar;

use strict;
use warnings;

use Net::NATS::Client;
use YAML::XS;
use File::Spec;
use File::Find;
use File::Basename;
use File::Slurp;
use File::Copy;
use File::Path;
use MIME::Base64;
use F5::ProtocolBuffers;
use JSON;
use POSIX;
use Digest::MD5;
use Scalar::Util qw(reftype);

use F5::DateTimeUtils;
use F5::Deploy;
use F5::ASMConfig::GuaranteedClient;
use F5::ConfigSet;
use F5::SetActive;
use F5::Db::Single;
use F5::Log::Job;
use F5::DeviceConfig;
use F5::SystemUtils;
use F5::Captcha;
use F5::Utils::File;
use F5::GlobalSettings;
use F5::ClientSideJS;
use F5::PrepareConf::Base;
use F5::ImportExportPolicy::YAML;
use F5::BehavioralDos;
use F5::ObjUtils;

my $IP_INTELLIGENCE_DIR = F5::Deploy::get_path_by_type("ip_intelligence");
our $BD_CONFIG_DIR = F5::Deploy::get_path_by_type('bd_config_tables');

our $are_protobufs_loaded = 0;

our $icb_heartbeat_message_type = 'declTmm::dpcMessageV2';

our %HANDLE_ELEMENTS = map { ($_ => 1) }
  qw(waf_policy
     file
     waf_remote_log_profile
     application_security
     virtual_server_profile
     profile_websecurity
     waf_stack_attributes
     application_attributes
     waf_settings
     license);

our %LOG_PROFILE_FILTER = (0 => 'illegal',
                           1 => 'illegal-including-staged-signatures',
                           2 => 'all',
                          );

our %EXCEPTION_MESSAGE_REASON = (
    0 => 'unspecified',
    1 => 'oversized',
    2 => 'delete',
);

our %DISPLAY_ACTION = (
                       'add_or_update' => 'add/update',
                       'delete' => 'delete',
                       'validate' => 'validate',
                      );

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;
    my $self = bless {}, $class;

    $self->init(%params);
    return $self;
}

#----------------------------------------------------------------------

sub init {
    my($self, %params) = @_;

    my $config_file = $params{config_file} || $self->{__last_config_file};
    $self->{__last_config_file} = $config_file;

    if ($config_file) {
        my $config = YAML::XS::LoadFile($config_file);
        for my $k (keys %$config) {
            $self->{$k} = $config->{$k};
        }
    }

    for my $k (keys %params) {
        $self->{$k} = $params{$k};
    }

    for my $k (keys %ENV) {
        if ($k =~ /^F5WAF_MBIP_SIDECAR_(.*)$/) {
            $self->{lc $1} = $ENV{$k};
        }
    }

    $self->{run_js_obfuscation} = defined $self->{run_js_obfuscation} ? $self->{run_js_obfuscation} : 1;
    $self->{generate_initial_captchas} = defined $self->{generate_initial_captchas} ? $self->{generate_initial_captchas} : 1;

    $self->startup();
}

#----------------------------------------------------------------------

sub startup {
    my($self) = @_;

    # Backup initial databases
    F5::DbUtils::copy_sqlite_databases(suffix => 'initial', overwrite_existing => 0);

    F5::Db::Single->init;
    $self->{dbh} = F5::Db::Single->instance();
    $self->{asmconfig_client} = F5::ASMConfig::GuaranteedClient::get_client(username => 'mbip_sidecar', dbh => $self->{dbh});
    $self->{config_set} = F5::ConfigSet->new(dbh => $self->{dbh},
                                             logger => $self->{logger},
                                             client => $self->{asmconfig_client});

    $self->{messages_handled_count} = 0;
    $self->{handled_entities_count} = 0;
    $self->{ignored_entities_count} = 0;
    $self->{log_profiles_by_name} = {};
    $self->{log_profiles_by_id} = {};
    $self->{application_security_objs} = {};

    my $device_config = F5::DeviceConfig->new(dbh => $self->{dbh}, logger => $self->{logger});
    $device_config->apply();

    $self->{captcha} = F5::Captcha->new(dbh => $self->{dbh},
                                        logger => $self->{logger},
                                        asmconfig_client => $self->{asmconfig_client});
    if ($self->{generate_initial_captchas}) {
        if (! $self->{captcha}->last_promoted_dir()) {
            $self->{captcha}->generate_captchas(generate_count => 1, out_dir_suffix => "_init", check_if_captcha_in_use => 1); # generate a single captcha on startup - blocking
            $self->{captcha}->promote_captchas();
        }
    }
    $self->{is_captcha_in_use} = 0;

    $self->{clientside_js} = F5::ClientSideJS->new(logger => $self->{logger});

    $self->{behavioral_dos} = F5::BehavioralDos->new(dbh => $self->{dbh},
                                                     logger => $self->{logger});
    $self->{bados_state} = {};
    $self->{bados_state}->{transaction_id} = 0;
    $self->{bados_state}->{last_sent_message} = '';

    if ($self->{initial_full_dir}) {
        $self->{initial_full_dir} = F5::Deploy::map_path($self->{initial_full_dir});
        $self->{source_full_dir} = F5::Deploy::map_path($self->{source_full_dir});
        $self->create_initial_full_dir;
    }

    #recreating bd_config_manifest to give ownership to non-root user
    if ($self->{recreate_bd_config_manifest}) {
        my $protobuf_manifest_file_basename = "bd_cfg_manifest.dat";
        my $protobuf_manifest_file = "$BD_CONFIG_DIR/$protobuf_manifest_file_basename";
        my $protobuf_manifest_file_temp = "$protobuf_manifest_file.temp";
        File::Copy::copy($protobuf_manifest_file,$protobuf_manifest_file_temp);
        unlink $protobuf_manifest_file;
        File::Copy::move($protobuf_manifest_file_temp,$protobuf_manifest_file);
    }

    if ($self->{init_marker_filename}) {
        F5::Utils::File::touch($self->{init_marker_filename});
    }

    $self->{generation} = 0;
    $self->{sequence} = 0;
    $self->{last_sub_id} = 0;
    $self->{entity_map} = {};
    $self->{entity_map_by_type} = {};
    $self->{applied_files} = {};
    $self->{queued_tasks} = [];
    $self->{previous_display_vs_names} = YAML::XS::Dump({});
    $self->{active_waf_policies} = {};
    $self->{last_icb_heartbeat_time} = 0;

    $self->load_state();
    $self->reinit();
}

#----------------------------------------------------------------------

sub reinit {
    my($self, %params) = @_;

    my $rv_proto = $self->load_protobuf_definitions();

    my $rv_nats = 1;
    if ($rv_proto && !$self->{nats_client}) {
        $rv_nats = $self->connect_to_nats();
    }


    return $rv_proto && $rv_nats;
}

#----------------------------------------------------------------------

sub reset_to_initial_state {
    my($self) = @_;

    F5::DbUtils::restore_sqlite_databases(suffix => 'initial');
    $self->startup();
}

#----------------------------------------------------------------------

sub connect_to_nats {
    my($self, %params) = @_;

    $self->{logger}->debug("Connecting to NATS") if $self->{logger};
    my $username_prefix = '';
    if ($self->{nats_username}) {
        $username_prefix = $self->{nats_username};
        if ($self->{nats_password}) {
            $username_prefix .= ":$self->{nats_password}";
        }
        $username_prefix .= '@';
    }

    my $connect_rv;
    eval {
        # CPAN module Net::NATS::Client uses module URI under the hood in order to parse the URI. The
        # username and password cannot be extracted if an unknown protocol is used. Thus, although the
        # actual protocol is NATS internal protocol, this is listed as "ftp" in order for URI parsing to
        # work correctly.
        my %socket_args;
        if ($self->{nats_connect_timeout}) {
            $socket_args{Timeout} = $self->{nats_connect_timeout};
        }
        if ($self->{ssl_ca_file}) {
            $socket_args{SSL_ca_file} = $self->{ssl_ca_file};
        }
        if ($self->{ssl_cert_file}) {
            $socket_args{SSL_cert_file} = $self->{ssl_cert_file};
        }
        if ($self->{ssl_key_file}) {
            $socket_args{SSL_key_file} = $self->{ssl_key_file};
        }
        if ($self->{ssl_version}) {
            $socket_args{SSL_version} = $self->{ssl_version};
        }
        if ($self->{ssl_cipher_list}) {
            $socket_args{SSL_cipher_list} = $self->{ssl_cipher_list};
        }
        $self->{nats_client} = Net::NATS::Client->new(uri => 'ftp://' . "$username_prefix$self->{nats_host}:$self->{nats_port}",
                                                      socket_args => \%socket_args,
                                                     );

        $connect_rv = $self->{nats_client}->connect();
    };
    if ($@ || ! $connect_rv) {
        my $error_msg = $@ || '';
        $error_msg =~ s| at /\S+ line \d+$||;
        if ($error_msg) {
            $error_msg = "($error_msg)";
        }
        $self->{logger}->error("Couldn't connect to NATS $error_msg") if $self->{logger};
        delete $self->{nats_client};
        return;
    }
    $self->{logger}->info("Connected to NATS") if $self->{logger};
    $self->{logger}->info("f5-asec-policy-compiler state changed to READY") if $self->{logger};

    my $icb_subscription = $self->{nats_client}->subscribe($self->{nats_subject_icb}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_protobuf_icb_message(message => $message->data);
                                                           };
                                                           # this error is likely to be an "ASMConfigException" object
                                                           if (my $err = $@) {
                                                               $self->{logger}->error("Failed to handle NATS message:") if $self->{logger};
                                                               $self->{logger}->error($err) if $self->{logger};
                                                           }
                                                  });

    my $jobproxy_subscription = $self->{nats_client}->subscribe($self->{nats_subject_jobproxy_subscribe}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_jobproxy_message(message => $message->data);
                                                           };
                                                           # this error is likely to be an "ASMConfigException" object
                                                           if (my $err = $@) {
                                                               $self->{logger}->error("Failed to handle NATS message:") if $self->{logger};
                                                               $self->{logger}->error($err) if $self->{logger};
                                                           }
                                                  });
    $self->send_jobproxy_cleanup;
    my $admd_subscription = $self->{nats_client}->subscribe($self->{nats_subject_admd_subscribe}, sub {
                                                           my ($message) = @_;
                                                           eval {
                                                               $self->handle_admd_message(message => $message->data);
                                                           };
                                                           # this error is likely to be an "ASMConfigException" object
                                                           if (my $err = $@) {
                                                               $self->{logger}->error("Failed to handle NATS message:") if $self->{logger};
                                                               $self->{logger}->error($err) if $self->{logger};
                                                           }
                                                  });

    return 1;
}

#----------------------------------------------------------------------

sub check_nats_connection {
    my($self) = @_;

    my $connection = $self->{nats_client}->connection();
    if ($connection) {
        if ($connection->eof()) {
            $self->{logger}->info("NATS connection was terminated") if $self->{logger};
            delete $self->{nats_client};
            return;
        }
        if (my $error_msg = $connection->error()) {
            $self->{logger}->error("NATS connection ended with error ($error_msg)") if $self->{logger};
            delete $self->{nats_client};
            return;
        }
    }
    else {
        $self->{logger}->info("NATS connection was lost") if $self->{logger};
        delete $self->{nats_client};
        return;
    }
    return 1;
}

#----------------------------------------------------------------------
# Protobuf defintions should only be loaded once for a process. Once .proto files from a particular
# path have been loaded, this will be intentionally skipped.

sub load_protobuf_definitions {
    my($self) = @_;

    return 1 if ($are_protobufs_loaded);

    if (! -d $self->{protobuf_source_path}) {
        $self->{logger}->info("Couldn't find protobuf defintions") if $self->{logger};
        return $are_protobufs_loaded;
    }

    $self->{logger}->debug("Loading protobuf defintions") if $self->{logger};

    $are_protobufs_loaded = F5::ProtocolBuffers::init(base_dir => $self->{protobuf_source_path},
                                                      package => $self->{protobuf_base_package},
                                                      imports_dir => $self->{protobuf_local_dir},
                                                     );

    if ($are_protobufs_loaded) {
        $self->{logger}->info("Loaded protobuf defintions") if $self->{logger};
    }

    return $are_protobufs_loaded;
}

#----------------------------------------------------------------------

sub send_bados_config {
    my($self) = @_;

    my @waf_virtual_server_profiles = $self->get_waf_virtual_server_profiles();

    my @bados_virtual_server_profiles;
    # Only send VS names currently in use by WAF policies
    for my $virtual_server_profile (@waf_virtual_server_profiles) {
        my $bados_vsp = {vs_name => $virtual_server_profile->{virtual_server},
                         profile_name => $self->{entity_map}->{$virtual_server_profile->{profile}}->{name}};
        push (@bados_virtual_server_profiles, $bados_vsp);
    }

    my $bados_config = $self->{behavioral_dos}->get_bados_config_message(virtual_server_profiles => \@bados_virtual_server_profiles,
                                                                         transaction_id => $self->{bados_state}->{transaction_id},
                                                                        );
    if ($bados_config eq $self->{bados_state}->{last_sent_message}) {
        $self->{logger}->debug("Bados configuration have not changed -- skipping sending config") if $self->{logger};
    }
    else {
        $bados_config = $self->{behavioral_dos}->get_bados_config_message(virtual_server_profiles => \@bados_virtual_server_profiles,
                                                                          transaction_id => $self->{bados_state}->{transaction_id} + 1,
                                                                         );
        $self->{logger}->debug("Sending Bados config") if $self->{logger};
        if ($self->{nats_client}) {
            $self->{nats_client}->publish($self->{nats_subject_admd_publish}, $bados_config);
        }
        else {
            $self->{logger}->debug("No NATS connection -- skipping send BADOS config") if $self->{logger};
        }
        $self->{bados_state}->{last_sent_message} = $bados_config;
        $self->{bados_state}->{transaction_id}++;
    }
}

#----------------------------------------------------------------------
# send_icb_heartbeat sends a request to ICB to receive all AsyncAPI messages starting from the
# current sequence (seq_start). Initially, seq_start is 0, in order to receive all messages from
# scratch.
#
# Once some AsyncAPI messages have been handled, seq_start will be the last sequence handled. This
# is sent periodically to ensure that new messages were not missed.
#
# To avoid unnecessary repeated requests to ICB, the timer to icb_heartbeat_interval is reset to the
# current time upon every successfully handled message. Conversely, a message that cannot be handled
# (such as due to NATS message truncation) will trigger a request for the current sequence.
#
# Note: "heartbeat" is somewhat of a misnomer, but the naming has been retained to maintain
# consistent logs.

sub send_icb_heartbeat {
    my($self) = @_;

    if (! $are_protobufs_loaded) {
        $self->{logger}->debug("Skipping ICB heartbeat (protobuf definitions are not loaded)") if $self->{logger};
        return;
    }

    if (! $self->{nats_client}) {
        $self->{logger}->debug("Skipping ICB heartbeat (NATS Client not connected)") if $self->{logger};
        return;
    }

    my $message = $icb_heartbeat_message_type->new();
    # default accessors names are: 'set_' + protobuf message field name
    $message->set_generation($self->{generation});
    $message->set_seq_start($self->{sequence});
    $self->{logger}->trace("Created $icb_heartbeat_message_type message:") if $self->{logger};
    $self->{logger}->trace($message) if $self->{logger};

    my $bytes = $icb_heartbeat_message_type->encode($message);

    if (! $self->{sent_initial_heartbeat_message}) {
        $self->{logger}->info("Sending initial heartbeat message") if $self->{logger};
        $self->{sent_initial_heartbeat_message} = 1;
    }

    $self->{nats_client}->publish($self->{heartbeat_subject}, $bytes);

    $self->{logger}->debug("Sent heartbeat message") if $self->{logger};
}

#----------------------------------------------------------------------

sub run {
    my($self) = @_;

    if ($self->{icb_message_load_filename}) {
        $YAML::XS::LoadBlessed = 1;
        my @messages = YAML::XS::LoadFile($self->{icb_message_load_filename});
        for my $message (@messages) {
            if ($message->{__message_type} && $message->{__message_type} eq 'job_proxy') {
                $self->handle_jobproxy_message(message => $message->{message});
            }
            else {
                $self->handle_icb_message(message_rec => $message);
            }
        }
        return;
    }

    my $last_queued_tasks_time = time;
    my $last_pulse_time = time;
    my $last_captcha_time = 0;
    $self->{last_clientside_js_obfuscation_time} = 0;

    $self->{logger}->info("f5-asec-policy-compiler state changed to STARTED") if $self->{logger};
    while (1) {
        $self->reinit();

        my @potential_poll_times;

        if ($self->{icb_heartbeat_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{icb_heartbeat_interval},
                                                                    start_time => $self->{last_icb_heartbeat_time}));
        }
        if ($self->{pulse_msg_interval}) {
            push(@potential_poll_times,
                 F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                                    start_time => $last_pulse_time));
        }

        my $poll_time = List::Util::min(grep { $_ } @potential_poll_times) || 5;
        if ($self->{nats_client}) {
            eval {
                $self->{logger}->trace("Going to poll for $poll_time seconds") if $self->{logger};
                my $rv = $self->{nats_client}->wait_for_op($poll_time);
                my $display_rv = $rv || 'undef';
                $self->{logger}->trace("After poll, rv = $display_rv") if $self->{logger};

                if (! $rv) {
                    $self->check_nats_connection();
                }
            };
            if ($@) {
                my $error_msg = $@ || '';
                $error_msg =~ s| at /\S+ line \d+$||;
                if ($error_msg) {
                    $error_msg = "($error_msg)";
                }
                $self->{logger}->error("Couldn't poll NATS server $error_msg") if $self->{logger};
                delete $self->{nats_client};
            }
        }
        else {
            $self->{logger}->trace("Sleeping for $poll_time seconds") if $self->{logger};
            sleep $poll_time;
        }

        if ($self->{queued_tasks_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{queued_tasks_interval},
                                                               start_time => $last_queued_tasks_time) == 0) {
            $self->run_queued_tasks();
            $last_queued_tasks_time = time;
        }

        if ($self->{pulse_msg_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{pulse_msg_interval},
                                                               start_time => $last_pulse_time) == 0) {
            $self->log_pulse_message;
            $last_pulse_time = time;
        }
        if ($self->{captcha_generate_interval} && $self->{is_captcha_in_use} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{captcha_generate_interval},
                                                               start_time => $last_captcha_time) == 0) {

            $self->spawn_generate_captchas();
            $last_captcha_time = time;
        }

        if ($self->{js_obfuscation_interval} && $self->{run_js_obfuscation} &&
            %{$self->{active_waf_policies}} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{js_obfuscation_interval},
                                                               start_time => $self->{last_clientside_js_obfuscation_time}) == 0) {
            $self->generate_clientside_js();
            $self->{last_clientside_js_obfuscation_time} = time;
        }
        if ($self->{icb_heartbeat_interval} &&
            F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{icb_heartbeat_interval},
                                                               start_time => $self->{last_icb_heartbeat_time}) == 0) {

            $self->send_icb_heartbeat();
            $self->{last_icb_heartbeat_time} = time;
        }

        if ($self->{current_captcha_pid}) {
            $self->reap_spawned_processes();
            $self->{captcha}->promote_captchas();
        }

        if ($self->{clientside_js}->check_obfuscated_files()) {
            my $prepare_conf = F5::PrepareConf::Base->new(dbh => $self->{dbh}, logger => $self->{logger});
            $prepare_conf->prepare_client_side_js();
            $self->set_active();
        }

    }
}

#----------------------------------------------------------------------

sub log_pulse_message {
    my($self) = @_;

    $self->{logger}->info("$self->{messages_handled_count} messages handled since startup ($self->{handled_entities_count} entities handled, $self->{ignored_entities_count} entities ignored)") if $self->{logger};
}

#----------------------------------------------------------------------

sub expand_proto_any_type_messages {
    my($message) = @_;

    if (ref($message) && ref($message) eq 'google.protobuf::Any') {
        $message = decode_single_any_type_message($message);
    }

    if (reftype($message) eq 'HASH') {
        for my $k (keys %$message) {
            if (ref($message->{$k})) {
                $message->{$k} = expand_proto_any_type_messages($message->{$k});
            }
        }
    }
    elsif (reftype($message) eq 'ARRAY') {
        for my $i (0..$#$message) {
            if (ref($message->[$i])) {
                $message->[$i] = expand_proto_any_type_messages($message->[$i]);
            }
        }
    }

    return $message;
}

#----------------------------------------------------------------------

sub decode_single_any_type_message {
    my($any_type_message) = @_;

    my $proto_type = $any_type_message->{type_url};
    $proto_type =~ s|.*/||;
    $proto_type =~ s|\.|::|;

    return $proto_type->decode($any_type_message->{value});
}

#----------------------------------------------------------------------

sub get_ref_suffix {
    my($struct) = @_;

    my $ref_suffix = '';
    if (ref($struct)) {
        $ref_suffix = ref($struct);
        $ref_suffix =~ s/^.*:://;
    }
    return $ref_suffix;
}

#----------------------------------------------------------------------

sub handle_admd_message {
    my($self, %params) = @_;
    my $message = $params{message};
    my $message_rec;
    eval {
        $message_rec = $self->{behavioral_dos}->deserialize_bados_config_message(message => $message);
    };
    if (my $err = $@) {
        $self->check_nats_connection();
        $err =~ s/\s+ at \s+ \S+ \s+ line \s+ \d+ .*$//sx;
        $self->{logger}->error("Couldn't decode BadosConfigRequest message ($err)") if $self->{logger};
        my $message_as_base64 = encode_base64($message);
        $self->{logger}->debug("Received message = \"$message_as_base64\"") if $self->{logger};
        return;
    }

    $self->{logger}->trace($message_rec) if $self->{logger};

    if (! $message_rec || ! %$message_rec ) {
        $self->{logger}->debug("Empty message -- skipping") if $self->{logger};
        return;
    }

    if (defined $message_rec->{last_received_transaction_id} && $message_rec->{last_received_transaction_id} < $self->{bados_state}->{transaction_id} && $self->{bados_state}->{last_sent_message}) {
        $self->{logger}->debug("Resending last Bados config") if $self->{logger};
        $self->{nats_client}->publish($self->{nats_subject_admd_publish}, $self->{bados_state}->{last_sent_message});
    }


}

#----------------------------------------------------------------------

sub handle_protobuf_icb_message {
    my($self, %params) = @_;
    my $message = $params{message};

    my $base_proto_type = $self->{protobuf_base_package} . '::' . $self->{protobuf_base_struct};
    my $message_rec;
    eval {
        $message_rec = $base_proto_type->decode($message);
        expand_proto_any_type_messages($message_rec);
    };
    if (my $err = $@) {
        $self->check_nats_connection();
        $err =~ s/\s+ at \s+ \S+ \s+ line \s+ \d+ .*$//sx;
        $self->{logger}->error("Couldn't decode $base_proto_type message ($err)") if $self->{logger};
        my $message_as_base64 = encode_base64($message);
        $self->{logger}->debug("Received message = \"$message_as_base64\"") if $self->{logger};
        return;
    }

    $self->handle_icb_message(%params,
                              message_rec => $message_rec);
}

#----------------------------------------------------------------------

sub handle_icb_message {
    my($self, %params) = @_;
    my $message_rec = $params{message_rec};

    $self->{logger}->trace($message_rec) if $self->{logger};

    if ($self->{icb_message_dump_filename}) {
        open(my $outf, '>>', $self->{icb_message_dump_filename});
        print $outf YAML::XS::Dump($message_rec);
        close $outf;
    }

    if (! $message_rec || ! %$message_rec || ! $message_rec->{embedded}) {
        $self->{logger}->debug("Empty message -- skipping") if $self->{logger};
        return;
    }
    # Top-level dataPlaneWrapperV2 only contains "embedded", which is the real content
    $message_rec = $message_rec->{embedded};

    my $message_type = get_ref_suffix($message_rec);

    if ($message_type eq 'unsatisfiableMsgV2' &&
        $message_rec->{generation} == $self->{generation} &&
        $message_rec->{seq_start} == $self->{sequence}) {

        $self->{logger}->info("Received unsatisfiable message -- resetting sequence") if $self->{logger};
        $self->reset_to_initial_state();
        return;
    }

    if ($message_type eq 'createMsgV2') {
        my %header_defaults = (generation => 0,
                               seq_end => 0,
                               seq_start => 0,
                               end_flag => 0,
                               sub_id => 0,
                              );
        my $header = $message_rec->{header} || {};
        $header = { %header_defaults, %$header };

        if ($header->{generation} != $self->{generation}) {
            $self->{logger}->info("Skipping message from different generation (generation = $header->{generation})") if $self->{logger};
            return;
        }

        if ($header->{seq_start} != $self->{sequence}) {
            $self->{logger}->debug("Skipping message from different sequence (seq_start = $header->{seq_start}, seq_end = $header->{seq_end})") if $self->{logger};
            return;
        }

        if ($header->{sub_id} && $header->{sub_id} < $self->{last_sub_id} + 1) {
            $self->{logger}->debug("Skipping already handled sub-message (current = $header->{sub_id}, last received = $self->{last_sub_id})") if $self->{logger};
            return;
        }

        if ($header->{sub_id} && $header->{sub_id} > $self->{last_sub_id} + 1) {
            $self->{logger}->info("Missing sub-message (current = $header->{sub_id}, last received = $self->{last_sub_id})  -- will request sequence restart in heartbeat (sequence = $self->{sequence})") if $self->{logger};
            $self->{last_icb_heartbeat_time} = 0;
            return;
        }

        for my $entity (@{$message_rec->{embedded}}) {
            my $action = 'add_or_update';
            my $entity_type = get_ref_suffix($entity);

            if ($entity_type eq 'exceptionObject') {
                my $reason = $EXCEPTION_MESSAGE_REASON{$entity->{reason}};

                if ($reason eq 'unspecified') {
                    $self->{logger}->info("Unspecified exception in message (seq_start = $header->{seq_start}, seq_end = $header->{seq_end}") if $self->{logger};
                    next;
                }
                if ($reason eq 'oversized') {
                    $self->{logger}->info("Oversized exception in message (seq_start = $header->{seq_start}, seq_end = $header->{seq_end}") if $self->{logger};
                    next;
                }

                if ($reason eq 'delete') {
                    # For some reason, ICB sends deleted entities as an exceptionObject with the "delete" reason and the entity (with only id populated) as
                    # "embedded" under the exceptionObject.
                    # For our purposes, this means that the current entity is the embedded message.
                    $action = 'delete';
                    $entity = $entity->{embedded};
                    $entity_type = get_ref_suffix($entity);
                }
            }

            if ($HANDLE_ELEMENTS{$entity_type}) {
                $self->queue_task(entity => $entity,
                                  entity_type => $entity_type,
                                  action => $action,
                                 );
                $self->run_queued_tasks();
                $self->{handled_entities_count}++;
            }
            else {
                $self->{logger}->debug("Skipping unhandled entity (type = $entity_type)") if $self->{logger};
                $self->{ignored_entities_count}++;
            }
        }

        $self->{messages_handled_count}++;
        $self->{last_icb_heartbeat_time} = time;
        if ($header->{end_flag} == 1) {
            $self->{generation} = $header->{generation} || 0;
            if (! $self->{sequence}) {
                $self->after_initial_load();
            }
            $self->{sequence} = $header->{seq_end} || 0;
            $self->{last_sub_id} = 0;
            $self->{logger}->debug("Updated current sequence to $self->{sequence}") if $self->{logger};
        }
        else {
            $self->{last_sub_id} = $header->{sub_id};
            $self->{logger}->debug("Updated current sub_id to $self->{last_sub_id}") if $self->{logger};
        }
        $self->save_state();
    }
}

#----------------------------------------------------------------------

sub queue_task {
    my($self, %params) = @_;
    my $entity = $params{entity};
    my $entity_type = $params{entity_type};
    my $action = $params{action};
    my $create_time = time;

    my $function_name = "${action}_${entity_type}";
    my $task = { entity => $entity,
                 entity_type => $params{entity_type} || '',
                 action => $action || 'add_or_update',
                 function_name => $function_name,
                 create_time => $create_time,
                 id => Digest::MD5::md5_hex($entity->{id} || '', $action, $create_time),
               };
    my $prerequisite_function_name = "check_prerequisite_${function_name}";
    if ($self->can($prerequisite_function_name)) {
        $task->{prerequisite_function_name} = $prerequisite_function_name;
    }
    $self->{logger}->debug("Enqueuing task to $action $entity_type (task id = $task->{id})") if $self->{logger};

    push(@{$self->{queued_tasks}}, $task);
}

#----------------------------------------------------------------------

sub run_task {
    my($self, %params) = @_;
    my $entity = $params{entity};
    my $entity_type = $params{entity_type};
    my $function_name = $params{function_name};
    my $action = $params{action};

    my $entity_id = $entity->{id} || '';
    my $entity_name = '';
    if ($entity->{name}) {
        $entity_name = $entity->{name};
    }
    elsif ($self->{entity_map}->{$entity_id}) {
        $entity_name = $self->{entity_map}->{$entity_id}->{name};
    }
    $entity_name ||= $entity_id;
    my $display_entity_id = '';
    if ($entity_name ne $entity_id) {
        $display_entity_id = " (id = $entity_id)";
    }

    $self->{logger}->info("Applying $DISPLAY_ACTION{$action} for $entity_type \"$entity_name\"$display_entity_id") if $self->{logger};


    my $job = $self->create_job(operation_id => $function_name,
                                entity => $entity,
                                entity_type => $entity_type,
                                entity_name => $entity_name,
                               );
    eval {
        $job->set_started();
        $self->$function_name(%$entity,
                              entity_type => $entity_type,
                              job => $job);
        $job->set_completed();
    };
    if ($@) {
        $job->set_failed();
    }
}

#----------------------------------------------------------------------

sub run_queued_tasks {
    my($self, %params) = @_;

    while (1) {
        my $was_task_run = 0;
        my $i = 0;
        while ($i <= $#{$self->{queued_tasks}}) {
            my $task = $self->{queued_tasks}->[$i];
            $self->{logger}->debug("Attempting to run task $task->{action} $task->{entity_type} (task id = $task->{id})") if $self->{logger};

            if ($task->{prerequisite_function_name}) {
                if ($task->{create_time} &&
                    F5::DateTimeUtils::get_time_until_maximum_interval(maximum_interval => $self->{max_wait_for_task_prerequisite_interval},
                                                                       start_time => $task->{create_time}) > 0) {

                    my $prerequisite_function_name = $task->{prerequisite_function_name};
                    my $is_prerequisite_fulfilled = $self->$prerequisite_function_name(%{$task->{entity}});
                    if ($is_prerequisite_fulfilled) {
                        $self->{logger}->debug("Prerequisite for $task->{action} $task->{entity_type} check succeeded") if $self->{logger};
                    }
                    else {
                        my $log_level = 'debug';
                        if (! $task->{_initial_missing_prerequisite_logged}) {
                            $log_level = 'info';
                            $task->{_initial_missing_prerequisite_logged} = 1;
                        }
                        $self->{logger}->$log_level("Prerequisite for $task->{action} $task->{entity_type} is missing -- deferring") if $self->{logger};
                        $i++;
                        next;
                    }
                }
                else {
                    $self->{logger}->debug("Skipping prerequisite check for $task->{action} $task->{entity_type}") if $self->{logger};
                }
            }
            my $rv = $self->run_task(%$task);
            $was_task_run = 1;
            splice(@{$self->{queued_tasks}}, $i, 1);
        }

        if (! $was_task_run) {
            last;
        }
    }
}

#----------------------------------------------------------------------

sub load_state {
    my($self) = @_;

    my $state_obj_row = F5::DbUtils::select_singleton(dbh => $self->{dbh}, table_name => 'PLC.MBIP_SIDECAR_STATE');
    if ($state_obj_row && $state_obj_row->{mbip_sidecar_state}) {
        my $state_obj = YAML::XS::Load($state_obj_row->{mbip_sidecar_state});
        for my $k (%$state_obj) {
            $self->{$k} = $state_obj->{$k};
        }

        $self->{logger}->info("Loading previous sidecar state (sequence = $self->{sequence})") if $self->{logger};
        $self->{logger}->trace($state_obj) if $self->{logger};
    }
    else {
        $self->{logger}->debug("No previous state to load") if $self->{logger};
        my $global_settings = F5::GlobalSettings->new(logger => $self->{logger},
                                                      client => $self->{asmconfig_client});
        $global_settings->regenerate_data_protection_configuration();
    }
}

#----------------------------------------------------------------------

sub save_state {
    my($self) = @_;

    my @state_fields = qw(generation
                          sequence
                          entity_map
                          entity_map_by_type
                          bados_state
                        );
    my $state_obj = {};
    for my $store_field (@state_fields) {
        $state_obj->{$store_field} = $self->{$store_field};
    }
    $self->{logger}->debug("Saving sidecar state") if $self->{logger};
    $self->{logger}->trace($state_obj) if $self->{logger};
    F5::DbUtils::upsert_singleton(dbh => $self->{dbh},
                                  table_name => "PLC.MBIP_SIDECAR_STATE",
                                  rec => {
                                          mbip_sidecar_state => YAML::XS::Dump($state_obj),
                                          last_updated => time(),
                                         }
                                 );
}

#----------------------------------------------------------------------

sub create_job {
    my($self, %params) = @_;
    my $operation_id = $params{operation_id};
    my $entity = $params{entity};
    my $entity_type = $params{entity_type};
    my $entity_name = $params{entity_name};

    my %job_attributes;
    my $get_attributes_func = "get_job_attributes_for_$entity_type";
    if ($self->can($get_attributes_func)) {
        %job_attributes = $self->$get_attributes_func(%params);

    }

    if ($entity_name && $entity_type) {
        my $rest_element_name = $entity_type;
        $rest_element_name =~ s/_(.)/\U$1/g;
        $job_attributes{$rest_element_name . "Name"} = $entity_name;
    }

    return F5::Log::Job->new(logger => $self->{logger},
                             operationId => $operation_id,
                             component => 'mbip_sidecar',
                             %job_attributes,
                            );
}

#----------------------------------------------------------------------

sub get_job_attributes_for_waf_policy {
    my($self, %params) = @_;
    my $entity = $params{entity};
    my $filename = $entity->{filename};

    if (! $filename) {
        if ($entity->{id} && $self->{entity_map}->{$entity->{id}}) {
            $filename = $self->{entity_map}->{$entity->{id}}->{filename};
        }
    }

    my %job_attributes;
    if ($filename) {
        my $import_content;
        my $uri = $self->get_file_path_map()->{$filename};
        if ($uri) {
            if ($uri =~ m|^/| && -e $uri) {
                $import_content = File::Slurp::read_file($uri);
            }
            else {
                $import_content = F5::Utils::File::get_file_from_link(link => $uri)
            }
        }
        elsif ($entity->{importContent}) {
            $import_content = $entity->{importContent};
        }

        if ($import_content) {
            $job_attributes{policySize} = length($import_content);
            $job_attributes{policyChecksum} = Digest::MD5->new->add($import_content)->hexdigest;
        }
    }

    return %job_attributes;
}

#----------------------------------------------------------------------

sub handle_jobproxy_message {
    my($self, %params) = @_;
    my $message = $params{message};

    if ($self->{icb_message_dump_filename}) {
        open(my $outf, '>>', $self->{icb_message_dump_filename});
        print $outf YAML::XS::Dump({ __message_type => 'job_proxy',
                                     message => $message});
        close $outf;
    }

    eval {
        $message = decode_json($message);
    };
    if ($@) {
        $self->{logger}->error("Cannot parse JobProxy message '$message': $@");
        return;
    }

    $self->{logger}->trace("Handling JobProxy message:") if $self->{logger};
    $self->{logger}->trace($message) if $self->{logger};

    my $jobproxy_role = $message->{Role} || "";
    $self->{logger}->debug("Received JobProxy message with role $jobproxy_role") if $self->{logger};
    if ($jobproxy_role && $jobproxy_role ne $self->{job_proxy_role}) {
        $self->{logger}->debug("Job role is not $self->{job_proxy_role}, skipping job.") if $self->{logger};
        return;
    }

    my $message_type = $message->{MessageType};
    if ($message_type eq "Broadcast") {
        $self->accept_jobproxy_job;
    }
    elsif ($message_type eq "Cleanup") {
        $self->send_jobproxy_cleanup;
    }
    elsif ($message_type eq "Content") {
        my $request;
        eval {
            $request = decode_json($message->{Request});
            $request->{job_proxy_id} = $message->{Id};
        };
        if ($@) {
            $self->{logger}->error("Cannot parse JobProxy message's request '$message->{Request}': $@");
            $self->forward_jobproxy_job_to_icb(id => $message->{Id});
            return;
        }

        my $waf_policies_path = '/waf-policies';
        if (index($message->{Uri}, $waf_policies_path) == 0) {
            $self->queue_task(entity => $request,
                              entity_type => "waf_policy",
                              action => "validate",
                              );
            $self->run_queued_tasks();
        }
        elsif ($request->{wafSettings}) {
            $self->queue_task(entity => $request,
                              entity_type => "waf_settings",
                              action => "validate",
                             );
            $self->run_queued_tasks();
        }
        else {
            $self->forward_jobproxy_job_to_icb(id => $message->{Id});
        }
    }
}

#----------------------------------------------------------------------

sub accept_jobproxy_job {
    my($self) = @_;
    my %message = (messageType => "Request",
                    role => $self->{job_proxy_role},
                   );
    $self->send_jobproxy_message(message => \%message);
}

#----------------------------------------------------------------------

sub send_jobproxy_cleanup {
    my($self) = @_;
    my %message = (messageType => "Cleanup",
                   role => $self->{job_proxy_role},
                   timeoutAfter => $self->{job_proxy_job_timeout},
                   deleteAfter => $self->{job_proxy_job_delete},
                  );
    $self->send_jobproxy_message(message => \%message);
}

#----------------------------------------------------------------------

sub forward_jobproxy_job_to_icb {
    my($self, %params) = @_;
    my $jobproxy_job_id = $params{id};
    my %message = (messageType => "Response",
                    id => $jobproxy_job_id,
                    status => "PENDING",
                    role => "AGENTCONFIG",
                   );
    $self->send_jobproxy_message(message => \%message);
}

#----------------------------------------------------------------------

sub set_jobproxy_job_status {
    my($self, %params) = @_;
    my $id = $params{id};
    my $status = $params{status};
    my $msg = $params{msg};
    my %message = (messageType => "Response",
                    id => $id,
                    status => $status,
                    role => $self->{job_proxy_role},
                    message => $msg,
                   );
    $self->send_jobproxy_message(message => \%message);

}

#----------------------------------------------------------------------

sub send_jobproxy_message {
    my($self, %params) = @_;
    my $message = $params{message};
    $self->{logger}->debug("Sending JobProxy message") if $self->{logger};
    $self->{logger}->trace($message) if $self->{logger};
    my $json_message = encode_json $message;
    if ($self->{nats_client}) {
        $self->{nats_client}->publish($self->{nats_subject_jobproxy_publish}, $json_message);
    }
    else {
        $self->{logger}->debug("No NATS connection -- skipping send") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub validate_waf_policy {
    my($self, %params) = @_;

    my $job_proxy_id = $params{job_proxy_id};
    my $import_policy_name = $params{name};

    $self->{logger}->info("Validating policy $import_policy_name") if $self->{logger};
    my $import_rv = $self->import_waf_policy(%params,
                                             force_reimport => 1,
                                            );
    if (!$import_rv->{completed_successfully}) {
        my $failure_msg = '';
        for my $msg (@{$import_rv->{msgs}}) {
            $failure_msg .= $msg->{msg} . '\n';
        }
        my $msg = {title => "WAF policy validation has failed",
                    detail => $failure_msg || "Failed to import policy $import_policy_name",
                    };
        $self->{logger}->error($msg->{detail}) if $self->{logger};
        $self->set_jobproxy_job_status(id => $job_proxy_id,
                                        msg => $msg,
                                        status => "FAILED",
                                        );
        return;
    }
    $self->forward_jobproxy_job_to_icb(id => $job_proxy_id);
}
#----------------------------------------------------------------------

sub delete_waf_policy {
    my($self, %params) = @_;
    my $api_uuid = $params{id};
    my $job = $params{job};

    my $policy_name;
    my $filename;
    if ($self->{entity_map}->{$api_uuid} && $self->{entity_map}->{$api_uuid}->{entity_type} eq 'waf_policy') {
        $policy_name = $self->{entity_map}->{$api_uuid}->{name};
        $filename = $self->{entity_map}->{$api_uuid}->{filename};
    }
    if (! $policy_name) {
        $self->{logger}->info("Policy not found -- skipping delete (api_uuid: $api_uuid)") if $self->{logger};
        return;
    }

    $self->{logger}->info("Deleting policy $policy_name (api_uuid: $api_uuid)") if $self->{logger};

    my $call_flags = F5::ASMConfig::CallFlags->new(job => $job);

    my $policy_criteria = F5::ASMConfig::Policy->new(policy_name => $policy_name);

    $self->{asmconfig_client}->delete_policy($self->{asmconfig_client}->{session},
                                             $call_flags,
                                             $policy_criteria);

    $self->delete_from_entity_map(%params);
    $self->remove_applied_file(filename => $filename);
}

#----------------------------------------------------------------------

sub import_waf_policy {
    my($self, %params) = @_;
    my $import_policy_name = $params{name};
    my $file_entity_name = $params{filename};
    my $force_reimport = $params{force_reimport};

    my $file_path_map = $self->get_file_path_map();
    my $filename = $file_path_map->{$file_entity_name};
    if ($filename) {
        $self->{logger}->debug("Found file for $file_entity_name (full path = $filename)") if $self->{logger};
    }
    else {
        $filename = $file_entity_name;
        $self->{logger}->info("File entity not found -- using direct filename $filename") if $self->{logger};
    }

    my $call_flags = F5::ASMConfig::CallFlags->new();

    my $is_temp_file;
    if (! $filename || ! $file_entity_name) {
        $is_temp_file = 1;
        my $fh;
        ($fh, $filename) = File::Temp::tempfile();
        print $fh $params{importContent};
        close $fh;
    }

    if (! $force_reimport && $self->was_file_already_applied(filename => $file_entity_name)) {
        $self->{logger}->info("Policy $file_entity_name already applied -- skipping") if $self->{logger};
        my $policy = F5::ASMConfig::Cache::get_cached_policy(dbh => $self->{dbh},
                                                             policy => {policy_name => $import_policy_name},
                                                            );
        return { completed_successfully => 1,
                 policy => $policy,
               };
    }

    if (! -f $filename || -z $filename) {
        my $failure_msg = "No such policy file or empty file at: $filename";
        my %import_rv = ( completed_successfully => 0,
                          msgs => [{msg => $failure_msg}],
                        );
        $self->{logger}->error($failure_msg) if $self->{logger};
        return \%import_rv;
    }

    $self->{logger}->info("Preparing importing for policy $import_policy_name")  if $self->{logger};

    my %import_file_path_map = map {$_ => "file://".$file_path_map->{$_}} keys %$file_path_map;
    my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(filename => $filename,
                                                                              policy_name => $import_policy_name,
                                                                              do_validation => 1,
                                                                              file_path_map => \%import_file_path_map,
                                                                              force_policy_name_flg => 1,
                                                                              overwrite => 1,
                                                                             );

    my $import_rv = $self->{asmconfig_client}->import_policy($self->{asmconfig_client}->{session},
                                                $call_flags,
                                                $import_export_properties);

    if ($import_rv && $import_rv->{completed_successfully}) {
        # The environment variable F5WAF_MBIP_SIDECAR_REQUEST_LOGGERS is read in init
        if ($self->{request_loggers}) {
            $self->add_or_update_request_loggers(policy => $import_rv->{policy});
        }
    }

    $self->save_applied_file(filename => $file_entity_name);
    if ($is_temp_file) {
        unlink $filename;
    }

    return $import_rv;
}

#----------------------------------------------------------------------

sub add_or_update_request_loggers {
    my($self, %params) = @_;
    my $policy = $params{policy};

    my $import_export_policy_yaml = F5::ImportExportPolicy::YAML->new(asm_config_handler => $self->{asmconfig_client});
    $self->{logger}->debug("Adding request logger to policy $policy->{policy_name}") if $self->{logger};
    my $request_loggers_json = decode_json $self->{request_loggers};
    for my $request_logger ( @$request_loggers_json ) {
        my $method;
        my $filter_string;
        if (exists $request_logger->{'$filter'} ) {
            $filter_string = $request_logger->{'$filter'};
            delete $request_logger->{'$filter'};
            $method = "PATCH";
        }
        else {
            $method = "POST";
        }
        $import_export_policy_yaml->send_rest_call(item      => $request_logger,
                                                   uri       => "/mgmt/tm/asm/policies/$policy->{rest_uuid}/request-loggers",
                                                   method    => $method,
                                                   '$filter' => $filter_string,);
    }

}

#----------------------------------------------------------------------

sub check_filename_exists {
    my($self, %params) = @_;
    my $filename = $params{filename};

    my $file_path_map = $self->get_file_path_map();
    if ($file_path_map->{$filename}) {
        if (-e $file_path_map->{$filename}) {
            $self->{logger}->trace("Found file reference for $filename ($file_path_map->{$filename})") if $self->{logger};
            return 1;
        }
        else {
            $self->{logger}->debug("File referent for $filename is missing ($file_path_map->{$filename})") if $self->{logger};
            $self->clean_file_entities();
            return;
        }
    }
    elsif (-e $filename) {
        $self->{logger}->trace("Found local file for $filename") if $self->{logger};
        return 1;
    }
    $self->{logger}->debug("Could not find local file or file reference for $filename") if $self->{logger};
    return;
}

#----------------------------------------------------------------------

sub check_prerequisite_validate_waf_policy {
    my($self, %params) = @_;
    return $self->check_filename_exists(%params);
}

#----------------------------------------------------------------------

sub check_prerequisite_add_or_update_waf_policy {
    my($self, %params) = @_;
    return $self->check_filename_exists(%params);
}

#----------------------------------------------------------------------

sub check_prerequisite_validate_waf_settings {
    my($self, %params) = @_;

    if (! $params{filename}) {
        return 1;
    }
    return $self->check_filename_exists(%params);
}

#----------------------------------------------------------------------

sub check_prerequisite_add_or_update_waf_settings {
    my($self, %params) = @_;

    if (! $params{filename}) {
        return 1;
    }
    return $self->check_filename_exists(%params);
}

#----------------------------------------------------------------------

sub add_to_entity_map {
    my($self, %params) = @_;
    my $entity = { %params };
    delete $entity->{job};

    $self->{entity_map}->{$entity->{id}} = $entity;
    $self->{entity_map_by_type}->{$entity->{entity_type}}->{$entity->{id}} = $entity;
}

#----------------------------------------------------------------------

sub delete_from_entity_map {
    my($self, %params) = @_;

    delete $self->{entity_map}->{$params{id}};
    delete $self->{entity_map_by_type}->{$params{entity_type}}->{$params{id}};
}

#----------------------------------------------------------------------

sub get_file_path_map {
    my($self) = @_;

    my $file_entities = $self->{entity_map_by_type}->{file} || {};
    my $file_path_map = { map { ($_->{name} => $_->{local_filename}) } values %$file_entities };
    return $file_path_map;
}

#----------------------------------------------------------------------

sub clean_file_entities {
    my($self) = @_;
    my $file_entities = $self->{entity_map_by_type}->{file} || {};

    $self->{logger}->trace("Cleaning file entities") if $self->{logger};
    for my $file (values %$file_entities) {
        if ($file->{local_filename} && ! -e $file->{local_filename}) {
            $self->{logger}->info("File referent $file->{name} is missing ($file->{local_filename}) -- removing") if $self->{logger};
            $self->delete_from_entity_map(%$file)
        }
    }
}

#----------------------------------------------------------------------

sub add_or_update_waf_policy {
    my($self, %params) = @_;
    my $api_uuid = $params{id};

    my $import_policy_name = $params{name};
    $self->add_to_entity_map(%params);

    $self->{logger}->info("Applying updated policy $import_policy_name (api_uuid: $api_uuid)") if $self->{logger};

    my $import_rv = $self->import_waf_policy(%params);
    my $policy = $import_rv->{policy};

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  wait_for_enforcer => 0,
                                                  use_daemon => 0,
                                                  policy_id => $policy->{policy_id},
                                                  update_ramcache => 0,
                                                  do_update_core => 1,
                                                  expanded_return_value => 1,
                                                  do_save_policy_history => 0,
                                                  asm_config_handler => $self->{asmconfig_client});
    if (!$set_active_rv->{completed_successfully}) {
        my $msg = "Failed to compile Policy $import_policy_name";
        $self->{logger}->error($msg)  if $self->{logger};
        die {"error_message" => $msg};
    }

    if (! $self->{is_captcha_in_use}) {
        $self->{is_captcha_in_use} = $self->{captcha}->is_captcha_in_use;
    }
    $self->send_bados_config();
}

#----------------------------------------------------------------------

sub add_or_update_file {
    my($self, %params) = @_;

    my $name = $params{name};
    my $uri  = $params{uri};
    my $local_filename = $uri;

    $local_filename =~ s|^file:/+|/|;
    $local_filename = "$self->{fcdn_base_dir}$local_filename";

    $self->{logger}->debug("Adding file $name (full path = $local_filename)") if $self->{logger};
    $self->add_to_entity_map(%params, local_filename => $local_filename);
    $self->clean_file_entities();
}

#----------------------------------------------------------------------

sub delete_file {
    my($self, %params) = @_;
    my $id = $params{id};
    my $entity = $self->{entity_map}->{$id} || {};
    my $name = $params{name} || $entity->{name};
    my $uri = $params{uri} || $entity->{uri} || '';

    if (! $name) {
        $self->{logger}->info("Could not find file to delete (id = $id)") if $self->{logger};
        return;
    }
    $self->{logger}->debug("Deleting file $name ($uri)") if $self->{logger};
    $self->delete_from_entity_map(%params);
    $self->clean_file_entities();
}

#----------------------------------------------------------------------
sub add_or_update_waf_remote_log_profile {
    my($self, %params) = @_;

    my $log_profile = \%params;
    $self->{logger}->info("Adding logging profile $log_profile->{name} ($log_profile->{id})") if $self->{logger};

    my @ip_address_segments = unpack('C*', pack('Q', $log_profile->{addr}->{a2}));
    $log_profile->{ip_address} = join('.', @ip_address_segments[4..7]);
    $self->{log_profiles_by_id}->{$log_profile->{id}} = $log_profile;
    $self->{log_profiles_by_name}->{$log_profile->{name}} = $log_profile;

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_waf_remote_log_profile {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting logging profile $params{name}") if $self->{logger};
    my $log_profile = $self->{log_profiles_by_name}->{$params{name}};
    delete $self->{log_profiles_by_name}->{$log_profile->{name}};
    delete $self->{log_profiles_by_id}->{$log_profile->{id}};

    $self->prepare_logging_profiles();
}

#----------------------------------------------------------------------

sub add_or_update_waf_stack_attributes {
    my($self, %params) = @_;

    $self->{logger}->info("Adding stack $params{stack_name} to application UUID $params{application_id} for virtual server UUID $params{virtual_server_id}") if $self->{logger};
    $self->add_to_entity_map(%params, name => $params{stack_name});
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub delete_waf_stack_attributes {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting stack from application UUID $self->{entity_map}->{$params{id}}->{application_id} for virtual server UUID $params{id}") if $self->{logger};
    $self->delete_from_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub add_or_update_application_attributes {
    my($self, %params) = @_;

    $self->{logger}->info("Adding application $params{name}") if $self->{logger};
    $self->add_to_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub delete_application_attributes {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting application $params{id}") if $self->{logger};
    $self->delete_from_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub add_or_update_profile_websecurity {
    my($self, %params) = @_;

    $self->add_to_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub delete_profile_websecurity {
    my($self, %params) = @_;

    $self->delete_from_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub add_or_update_virtual_server_profile {
    my($self, %params) = @_;

    $self->add_to_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub delete_virtual_server_profile {
    my($self, %params) = @_;

    $self->delete_from_entity_map(%params);
    $self->prepare_waf_stack_attributes();
}

#----------------------------------------------------------------------

sub set_active {
    my($self, %params) = @_;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  expanded_return_value => 1,
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client},
                                                  %params,
                                                 );

    if (!$set_active_rv->{completed_successfully}) {
        $self->{logger}->error("Failed to set active") if $self->{logger};
    }
}

#----------------------------------------------------------------------

sub get_waf_virtual_server_profiles {
    my($self) = @_;

    my @waf_virtual_server_profiles;
    for my $virtual_server_profile (values %{$self->{entity_map_by_type}->{virtual_server_profile}}) {
        if ($self->{entity_map}->{$virtual_server_profile->{profile}} &&
            ($self->{entity_map}->{$virtual_server_profile->{profile}}->{entity_type} eq 'waf_policy' ||
             $self->{entity_map}->{$virtual_server_profile->{profile}}->{entity_type} eq 'profile_websecurity')) {
            push (@waf_virtual_server_profiles, $virtual_server_profile);
        }
    }
    return @waf_virtual_server_profiles;
}

#----------------------------------------------------------------------

sub prepare_waf_stack_attributes {
    my ($self, %params) = @_;

    my $dbh    = $params{dbh} || $self->{dbh};
    my $logger = $params{logger} || $self->{logger};
    my $display_vs_names = {};
    my $active_waf_policies = {};

    $self->send_bados_config();

    # waf_stack_attributes map by virtual_server_id
    my $waf_stack_attributes_map = { map { ($_->{virtual_server_id} => $_) } grep { $_->{virtual_server_id} } values %{$self->{entity_map_by_type}->{waf_stack_attributes}} };

    # Only map VS names currently in use by WAF policies
    my @waf_virtual_server_profiles = $self->get_waf_virtual_server_profiles();
    for my $virtual_server_profile (@waf_virtual_server_profiles) {
        my $profile_websecurity = $self->{entity_map}->{$virtual_server_profile->{profile}};
        $active_waf_policies->{$profile_websecurity->{id}} = $profile_websecurity;

        my $waf_stack_attributes = $waf_stack_attributes_map->{$virtual_server_profile->{virtual_server}};

        if ($waf_stack_attributes) {
            $self->{logger}->trace("Found this waf_stack_attributes entity:") if $self->{logger};
            $self->{logger}->trace($waf_stack_attributes) if $self->{logger};
            my $application = $self->{entity_map}->{$waf_stack_attributes->{application_id}};
            if ($application) {
                my $entity_display_vs_name = "$application->{name}/$waf_stack_attributes->{stack_name}";
                $display_vs_names->{$virtual_server_profile->{virtual_server}} = { display_vs_name => $entity_display_vs_name};
            }
        }
    }
    $self->{active_waf_policies} = $active_waf_policies;

    my $dump_display_vs_names = YAML::XS::Dump($display_vs_names);
    if ($dump_display_vs_names eq $self->{previous_display_vs_names}) {
        $self->{logger}->debug("Stack attributes have not changed -- skipping set active") if $self->{logger};
    }
    else {
        $self->{logger}->debug("Updating stack attributes") if $self->{logger};
        my $prepare_conf = F5::PrepareConf::Base->new(dbh => $dbh,
                                                      logger => $logger,
                                                     );
        $prepare_conf->prepare_display_vs_name(display_vs_names => $display_vs_names);
        $self->set_active(%params);
        $self->{previous_display_vs_names} = $dump_display_vs_names;
    }
}

#----------------------------------------------------------------------

sub add_or_update_application_security {
    my($self, %params) = @_;

    my $waf_log_profile_id = $params{wafLogProfileId} || '';
    $self->{logger}->info("Adding application security $params{name} (virtualServerId = $params{virtualServerId}, wafLogProfileId = $waf_log_profile_id)") if $self->{logger};
    $self->{application_security_objs}->{$params{name}} = { %params };

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub delete_application_security {
    my($self, %params) = @_;

    $self->{logger}->info("Deleting application security $params{name}") if $self->{logger};
    delete $self->{application_security_objs}->{$params{name}};

    $self->prepare_logging_profiles(%params);
}

#----------------------------------------------------------------------

sub prepare_logging_profiles {
    my($self, %params) = @_;

    $self->{logger}->debug("Updating logging profiles") if $self->{logger};

    my @vs_configs;
    for my $security_obj (values %{$self->{application_security_objs}}) {
        if (! $security_obj->{wafLogProfileId}) {
            $self->{logger}->info("WAF Security $security_obj->{name} has no logging profile -- skipping") if $self->{logger};
            next;
        }

        my $log_profile = $self->{log_profiles_by_id}->{$security_obj->{wafLogProfileId}};
        if (! $log_profile) {
            $self->{logger}->info("WAF Log Profile $security_obj->{wafLogProfileId} doesn't exist -- skipping") if $self->{logger};
            next;
        }
        my $logging_config = { logging => {
                                           dest_ip => $log_profile->{ip_address},
                                           dest_port => $log_profile->{port},
                                           filter => $LOG_PROFILE_FILTER{$log_profile->{filterRequestType}},
                                           content => {
                                                       max_message_size => '64k',
                                                      },
                                          },
                             };

        my $vs_config = $self->{config_set}->build_logging_config(policy => { vs_info => { $security_obj->{virtualServerId} => $logging_config } });
        push(@vs_configs, @$vs_config);
    }

    $self->{config_set}->prepare_all_logging_profiles(vs_config => \@vs_configs);
    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  job => $params{job},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});
}

#----------------------------------------------------------------------

sub create_initial_full_dir {
    my($self) = @_;

    my $set_active_rv = F5::SetActive::set_active(dbh => $self->{dbh},
                                                  logger => $self->{logger},
                                                  do_update_core => 1,
                                                  asm_config_handler => $self->{asmconfig_client});

    mkpath($self->{initial_full_dir});
    find({
          no_chdir => 1,
          follow_fast => 1,
          wanted => sub {
              my $filename = basename($File::Find::name);
              copy($File::Find::name, "$self->{initial_full_dir}/$filename");
          },
         },
         $self->{source_full_dir});
}

#----------------------------------------------------------------------

sub spawn_generate_captchas {
    my($self, %params) = @_;

    if ($self->{current_captcha_pid} = fork) {
        # in parent
        $self->{logger}->debug("Forked to the parent with child pid $self->{current_captcha_pid}") if $self->{logger};
    }
    else {
        # in child
        exit($self->{captcha}->generate_captchas(%params) || 0);
    }
}

#----------------------------------------------------------------------

sub reap_spawned_processes {
    my($self, %params) = @_;

    if ($self->{current_captcha_pid}) {
        my $rv = waitpid($self->{current_captcha_pid}, WNOHANG);
        if ($rv) {
            my $exit_code = $? >> 8;
            $self->{logger}->debug("Generate captchas: child exited with code: $exit_code") if $self->{logger};
            delete $self->{current_captcha_pid};
        }
    }
}

#----------------------------------------------------------------------

sub delete_waf_settings {
    my($self, %params) = @_;

    $self->delete_from_entity_map(%params);
}

#----------------------------------------------------------------------

sub add_or_update_waf_settings {
    my($self, %params) = @_;

    $self->apply_waf_settings(%params);
    $self->add_to_entity_map(%params);
}

#----------------------------------------------------------------------

sub validate_waf_settings {
    my($self, %params) = @_;

    my $job_proxy_id = $params{job_proxy_id};
    my $waf_settings = $params{wafSettings} || {};
    if ($params{filename}) {
        $waf_settings->{filename} = $params{filename};
    }
    eval {
        $self->apply_waf_settings(%$waf_settings);
    };

    if ($@) {
        my $msg = $@;
        $self->set_jobproxy_job_status(id => $job_proxy_id,
                                       msg => $msg,
                                       status => "FAILED",
                                      );
        die $msg;
    }

    $self->forward_jobproxy_job_to_icb(id => $job_proxy_id);
}

#----------------------------------------------------------------------

sub apply_waf_settings {
    my($self, %params) = @_;

    my $waf_settings = { %params };
    $waf_settings = F5::ObjUtils::convert_to_deep_hash($waf_settings);
    my $global_settings;

    eval {
        $global_settings = F5::GlobalSettings->new(%$waf_settings,
                                                   logger => $self->{logger},
                                                   client => $self->{asmconfig_client},
                                                   file_path_map => $self->get_file_path_map(),
                                                   allow_cookie_protection_seed_from_file => $self->{allow_cookie_protection_seed_from_file},
                                                  );
    };
    if ($@) {
        my $err = $@;
        my $msg = {title => "Failed to load WAF settings file",
                   detail => $err,
                  };
        die $msg;
    }

    my $apply_rv = $global_settings->apply();
    if (defined $apply_rv->{completed_successfully} && !$apply_rv->{completed_successfully}) {
        my $msg = {title => "Failed to apply WAF settings",
                   detail => $apply_rv->{failure_msg},
                  };
        die $msg;
    }
}

#----------------------------------------------------------------------

sub add_or_update_license {
    my($self, %params) = @_;

    my %ip_intelligence_config = (
        'ip-intelligence' => {
            'licenseKey' => $params{clusteruuid},
        }
    );
    YAML::XS::DumpFile("$IP_INTELLIGENCE_DIR/ip_intelligence_config.yaml", \%ip_intelligence_config);
}

#----------------------------------------------------------------------

sub after_initial_load {
    my($self) = @_;

}

#----------------------------------------------------------------------

sub generate_clientside_js {
    my($self, %params) = @_;

    $self->{logger}->debug("Generating client-side JS") if $self->{logger};
    $self->{clientside_js}->generate_obfuscation_config_file(generated_variants_count => $self->{js_obfuscation_copies});
}

#----------------------------------------------------------------------

sub was_file_already_applied {
    my($self, %params) = @_;
    my $filename = $params{filename};

    $self->{logger}->debug("Checking whether file has already been applied (filename = $filename)") if $self->{logger};
    my $new_applied_file = $self->get_applied_file_rec(filename => $filename,
                                                       file_path_map => $self->get_file_path_map(),
                                                      );

    if (! $self->{applied_files}->{$filename}) {
        $self->{logger}->debug("File $filename was not yet applied") if $self->{logger};
        return;
    }
    my $old_applied_file = $self->{applied_files}->{$filename};
    my $does_file_differ = 0;
    for my $k (keys %$old_applied_file) {
        if ($old_applied_file->{$k} ne $new_applied_file->{$k}) {
            $self->{logger}->debug("File $filename differs by attribute $k (old value = $old_applied_file->{$k}, new value = $new_applied_file->{$k})") if $self->{logger};
            $does_file_differ = 1;
        }
    }
    if ($does_file_differ) {
        $self->{logger}->info("Previously applied file $filename differs and must be re-applied") if $self->{logger};
        return;
    }

    return 1;
}

#----------------------------------------------------------------------

sub save_applied_file {
    my($self, %params) = @_;
    my $filename = $params{filename};

    my $applied_file = $self->get_applied_file_rec(filename => $filename,
                                                   file_path_map => $self->get_file_path_map(),
                                                  );

    $self->{applied_files}->{$filename} = $applied_file;
}

#----------------------------------------------------------------------

sub remove_applied_file {
    my($self, %params) = @_;
    my $filename = $params{filename};

    if ($self->{applied_files}->{$filename}) {
        delete $self->{applied_files}->{$filename};
    }
}

#----------------------------------------------------------------------

sub get_applied_file_rec {
    my($self, %params) = @_;
    my $filename = $params{filename};

    my $actual_filename = $self->get_file_path_map()->{$filename};
    if (! $actual_filename) {
        $actual_filename = $filename;
    }
    my $applied_file = {
                        filename => $filename,
                        actual_filename => $actual_filename,
                        file_size => -s $actual_filename,
                        file_md5 => F5::Utils::File::get_file_md5($actual_filename),
                       };
}

#----------------------------------------------------------------------

1;
