package F5::Policy::Convert;

use strict;
use warnings;

use JSON;
use File::Spec;
use Storable qw(fd_retrieve);
use YAML::XS;

use F5::ASMConfig::GuaranteedClient;
use F5::Utils::Storable;
use F5::ExportPolicy::LegacyBinary;
use F5::Deploy;
use F5::Utils::File;
use F5::BigipUtils;
use F5::Logger;

#----------------------------------------------------------------------

sub new {
    my($class, %params) = @_;

    my $self = {};
    bless $self, $class;
    $self->init(%params);

    return $self;
}
#----------------------------------------------------------------------

sub init {
    my ($self, %params) = @_;

    $self->{logger} = $params{logger} || F5::Logger->new('/ts/log/policy_convert.log');
    $self->{client} = $params{client};
    if (! $self->{client}) {
        $self->{client} = F5::ASMConfig::GuaranteedClient::get_client((username => 'import_export_policy'));
    }

    $self->{temporary_files} = [];
}

#----------------------------------------------------------------------

sub convert_policy {
    my($self, %params) = @_;

    my $import_filename = $params{import_file};
    my $export_filename = $params{export_file};
    my $format = lc($params{format});
    my $keep_full_configuration = $params{keep_full_configuration};
    my $template = $params{template};
    my $full_export = $params{full_export};
    my $perl32 = $params{perl32};

    my $result = {completed_successfully => 0};
    $self->{client}->begin_work();
    eval {
        my $in_filename  = File::Spec->rel2abs($import_filename);
        my $out_filename = File::Spec->rel2abs($export_filename);

        $self->{logger}->debug("Convert input file '$in_filename' to output file '$out_filename'") if $self->{logger};

        my $original_format = $format;
        if ($original_format eq 'binary32') {
            $format = 'binary';
            if ( ! -f $perl32 ) {
                die "perl32 '$perl32' not found\n";
            }
        }
        elsif ($format eq 'yaml') {
            die "Export to YAML format is not supported\n";
        }

        if (F5::Utils::File::is_file_gzip_storable32(filename => $in_filename)) {
            $self->{logger}->info("Convert storable32 to YAML from file: $in_filename") if $self->{logger};
            $in_filename = $self->convert_storable32_to_yaml(in_filename => $in_filename);
        }

        my $call_flags = F5::ASMConfig::CallFlags->new();
        my $import_export_properties = F5::ASMConfig::ImportExportProperties->new(
                                                                                  filename => $in_filename,
                                                                                  enforce_strictness => ($keep_full_configuration ? 0 : 1),
        );
        if ($keep_full_configuration) {
            #Allow all features during import
            $ENV{F5WAF_ENABLE_ALL_FEATURES} = 1;
        }

        ## Import Policy
        $result = $self->{client}->import_policy($self->{client}->{session},
                                         $call_flags,
                                         $import_export_properties);
        return $result unless $result->{completed_successfully};

        # Messages should be taken from the import for JSON policy conversion
        my $import_msgs;
        if ($result->{file_format} eq 'json' || $result->{file_format} eq 'yaml') {
            $import_msgs = $result->{msgs};
        }

        my $default_template = F5::BigipUtils::get_feature_value('default_template') || 'POLICY_TEMPLATE_BLANK';
        my $policy_template = uc $template || $default_template;
        $policy_template =~ s/^POLICY_TEMPLATE_//;
        my $icontrol_template_name = "POLICY_TEMPLATE_$policy_template";
        $policy_template = F5::ASMConfig::PolicyTemplate->new(icontrol_name => $icontrol_template_name);

        $self->{logger}->debug("Export policy with template: $icontrol_template_name") if $self->{logger};
        ## Export Policy
        $result = $self->{client}->export_policy($self->{client}->{session},
                                         $call_flags,
                                         $result->{policy},
                                         F5::ASMConfig::ImportExportProperties->new(
                                                                                    filename => $out_filename,
                                                                                    file_format => $format,
                                                                                    policy_template => $policy_template,
                                                                                    enforce_strictness => ($keep_full_configuration ? 0 : 1),
                                                                                    minimal_export => ($full_export ? 0 : 1),
                                                                                   ));

        if ($original_format eq 'binary32') {
            $self->convert_to_binary32(perl32 => $perl32, result => $result, out_filename => $out_filename,);
        }

        if ($import_msgs) {
            $result->{msgs} ||= [];
            push (@{$result->{msgs}}, @$import_msgs);
        }
    };

    if ($@) {
        $result->{error_message} = ref $@ eq 'F5::ASMConfig::ASMConfigException' ? $@->{error_message} : $@;
        $result->{error_message} =~ s/\n$//;
        $result->{completed_successfully} = 0;
    }
    else {
        if ($result->{msgs}) {
            $result->{warnings} = [map { $_->{msg} } grep {$_->{level} eq 'warning'} @{$result->{msgs}}];
            $result->{error_message} = join("\n", map { $_->{msg} } grep {$_->{level} eq 'fatal'} @{$result->{msgs}});
        }
        if ($result->{filename} && $result->{file_size}) {
            $result->{completed_successfully} = 1;
        }
    }

    $self->{client}->rollback();

    $self->clean_temporary_files();
    return $result;
}

#----------------------------------------------------------------------

sub convert_to_binary32 {
    # legacy support for environments with "perl32" installed
    my($self, %params) = @_;

    my $perl32 = $params{perl32};
    my $result = $params{result};
    my $out_filename = $params{out_filename};

    my $yaml_file = File::Spec->rel2abs('policy.yaml');

    F5::Utils::Storable::read_file(in_filename => $out_filename, out_filename => $yaml_file);
    F5::ExportPolicy::LegacyBinary::convert_for_version(policy_file => $yaml_file, version => '15.1.0');

    my $yaml_to_storable_tool = F5::Deploy::map_path('/usr/share/ts/bin/convert_yaml_to_storable');

    my $rv = system($perl32, $yaml_to_storable_tool, $yaml_file, $out_filename);
    if ($rv) {
        $result->{completed_successfully} = 0;
        undef $result->{file_size};
    }
    else {
        $result->{file_size} = -s $out_filename;
    }
    unlink $yaml_file;
}

#----------------------------------------------------------------------

sub clean_temporary_files {
    my($self) = @_;

    for my $temp_file (@{$self->{temporary_files}}) {
        if (-e $temp_file) {
            unlink $temp_file;
        }
    }
}

#----------------------------------------------------------------------

sub convert_storable32_to_yaml {
    my($self, %params) = @_;

    my $in_filename = $params{in_filename};

    my $tmp_path = F5::Deploy::get_path_by_type('tmp');
    my $timestamp = time();
    my $yaml_filename = "$tmp_path/convert_storable_". $timestamp. ".yaml";

    my $python_exe = F5::SystemUtils::get_python_exe();
    my $convert2yaml = F5::Deploy::map_path('/usr/share/ts/bin/convert-storable-to-yaml');
    my $res = F5::SystemUtils::run_system_tool(command_line_params => [$python_exe,
                                                                       $convert2yaml,
                                                                       '-i',
                                                                       $in_filename,
                                                                       '-o',
                                                                       $yaml_filename]);

    if (! $res->{completed_successfully}) {
        die "Failed to convert storable32 binary policy to YAML file: ". $res->{standard_error}. "\n";
    }

    push @{$self->{temporary_files}}, $yaml_filename;

    my $binary64_filename = "$tmp_path/convert_storable_binary_". $timestamp. ".plc";
    F5::Utils::Storable::write_file(input_file => $yaml_filename, output_file => $binary64_filename);
    push @{$self->{temporary_files}}, $binary64_filename;

    return $binary64_filename;
}

#----------------------------------------------------------------------


1;

